from pathlib import Path
from typing import Any, Callable, Collection, Dict, List, Optional, Set

import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

from negmas import (
    AgentMechanismInterface,
    AspirationNegotiator,
    Issue,
    MappingUtilityFunction,
    MechanismState,
    Negotiator,
    RenegotiationRequest,
    SAOMechanism,
)
from negmas.events import Event, EventSink, EventSource
from negmas.helpers import unique_name
from negmas.situated import Action, Agent, Breach, Contract, World

results = []  # will keep results not to use printing


class DummyWorld(World):
    def complete_contract_execution(
        self, contract: Contract, breaches: List[Breach], resolved: bool
    ) -> None:
        pass

    def get_contract_finalization_time(self, contract: Contract) -> int:
        return self.current_step + 1

    def get_contract_execution_time(self, contract: Contract) -> int:
        return self.current_step

    def contract_size(self, contract: Contract) -> float:
        return 0.0

    def __init__(self, n_steps=10000, negotiation_speed=20, **kwargs):
        super().__init__(
            n_steps=n_steps,
            negotiation_speed=negotiation_speed,
            neg_n_steps=10,
            neg_time_limit=10,
            **kwargs,
        )
        self.the_agents = []

    def join(self, x: "Agent", simulation_priority: int = 0):
        super().join(x=x, simulation_priority=simulation_priority)
        self.the_agents.append(x)

    def delete_executed_contracts(self) -> None:
        pass

    def executable_contracts(self) -> Collection[Contract]:
        return []

    def post_step_stats(self):
        pass

    def pre_step_stats(self):
        pass

    def order_contracts_for_execution(
        self, contracts: Collection[Contract]
    ) -> Collection[Contract]:
        return contracts

    def contract_record(self, contract: Contract) -> Dict[str, Any]:
        return contract.__dict__

    def start_contract_execution(self, contract: Contract) -> Set[Breach]:
        return set()

    def _process_breach(
        self, contract: Contract, breaches: List[Breach], force_immediate_signing=True
    ) -> Optional[Contract]:
        return None

    def breach_record(self, breach: Breach):
        return breach.__dict__

    def execute_action(
        self, action: Action, agent: "Agent", callback: Callable = None
    ) -> bool:
        return True

    def get_private_state(self, agent: "Agent") -> Any:
        s = {"partners": [_ for _ in self.the_agents if _ is not agent]}
        return s

    def simulation_step(self, stage: int):
        pass


class DummyAgent(Agent):
    def init(self):
        pass

    def _respond_to_negotiation_request(
        self,
        initiator: str,
        partners: List[str],
        issues: List[Issue],
        annotation: Dict[str, Any],
        mechanism: AgentMechanismInterface,
        role: Optional[str],
        req_id: Optional[str],
    ) -> Optional[Negotiator]:
        negotiator = AspirationNegotiator(
            ufun=MappingUtilityFunction(mapping=lambda x: 1.0 - x[0] / 10.0)
        )
        return negotiator

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_failure(
        self,
        partners: List[str],
        annotation: Dict[str, Any],
        mechanism: AgentMechanismInterface,
        state: MechanismState,
    ) -> None:
        pass

    def on_negotiation_success(
        self, contract: Contract, mechanism: AgentMechanismInterface
    ) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        return self.id

    def on_contract_executed(self, contract: Contract) -> None:
        pass

    def on_contract_breached(
        self, contract: Contract, breaches: List[Breach], resolution: Optional[Contract]
    ) -> None:
        pass

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: List[Breach]
    ) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: List[Breach], agenda: RenegotiationRequest
    ) -> Optional[Negotiator]:
        return None

    def __init__(self, name=None):
        super().__init__(name=name)
        self.id = name
        self.__current_step = 0

    def step(self):
        global results
        if self.awi is None:
            return
        self.__current_step = self.awi.current_step
        if (self.__current_step == 2 and self.name.endswith("1")) or (
            self.__current_step == 4 and self.name.endswith("2")
        ):
            issues = [Issue(10, name="i1")]
            partners = self.awi.state["partners"]
            self._request_negotiation(
                partners=[_.name for _ in partners] + [self.name], issues=issues
            )
            results.append(f"{self.name} started negotiation with {partners[0].name}")
        results.append(f"{self.name}: step {self.__current_step}")


def test_world_runs_with_some_negs(capsys):
    global results
    results = []
    world = DummyWorld(n_steps=10)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    world.run()
    assert "A1: step 1" in results, "first step logged"
    assert "A2: step 1" in results, "first step logged"
    assert "A1 started negotiation with A2" in results, "negotiation started"
    assert f"A1: step {world.n_steps-1}" in results, "last step logged"
    assert f"A2: step {world.n_steps-1}" in results, "last step logged"
    assert len(world.saved_contracts) == 2
    assert sum(world.stats["n_negotiations"]) == 2


def test_config_reader_with_a_world():

    world = DummyWorld()
    assert world.bulletin_board is not None
    assert world.n_steps == 10000

    world = DummyWorld.from_config(scope=globals(), config={"n_steps": 10})
    assert world.bulletin_board is not None
    assert world.n_steps == 10


def test_config_reader_with_a_world_with_enum():
    world = DummyWorld()
    assert world.bulletin_board is not None
    assert world.n_steps == 10000

    world = DummyWorld.from_config(
        scope=globals(), config={"n_steps": 10, "negotiation_speed": 2}
    )
    assert world.bulletin_board is not None
    assert world.n_steps == 10
    assert world.negotiation_speed == 2


def test_world_picklable(tmp_path):
    import pickle

    world = DummyWorld()
    world.step()
    world.step()
    file = tmp_path / "world.pckl"

    with open(file, "wb") as f:
        pickle.dump(world, f)
    with open(file, "rb") as f:
        w = pickle.load(f)
    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    w.run()


@settings(deadline=None)
@given(
    single_checkpoint=st.booleans(),
    checkpoint_every=st.integers(0, 6),
    exist_ok=st.booleans(),
)
def test_world_auto_checkpoint(tmp_path, single_checkpoint, checkpoint_every, exist_ok):
    import shutil

    new_folder: Path = tmp_path / unique_name("empty", sep="")
    new_folder.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(new_folder)
    new_folder.mkdir(parents=True, exist_ok=True)
    filename = "mechanism"
    n_steps = 20

    world = DummyWorld(
        n_steps=n_steps,
        checkpoint_every=checkpoint_every,
        checkpoint_folder=new_folder,
        checkpoint_filename=filename,
        extra_checkpoint_info=None,
        exist_ok=exist_ok,
        single_checkpoint=single_checkpoint,
    )

    world.run()

    if 0 < checkpoint_every <= n_steps:
        if single_checkpoint:
            assert len(list(new_folder.glob("*"))) == 2, print(
                f"World ran for: {world.current_step}"
            )
        else:
            assert len(list(new_folder.glob("*"))) >= 2 * (
                max(1, world.current_step // checkpoint_every)
            )
    elif checkpoint_every > n_steps:
        assert len(list(new_folder.glob("*"))) == 2
    else:
        assert len(list(new_folder.glob("*"))) == 0


def test_world_checkpoint(tmp_path):
    world = DummyWorld()
    world.step()
    world.step()

    file_name = world.checkpoint(tmp_path)

    info = DummyWorld.checkpoint_info(file_name)
    assert isinstance(info["time"], str)
    assert info["step"] == 2
    assert info["type"].endswith("DummyWorld")
    assert info["id"] == world.id
    assert info["name"] == world.name

    w = DummyWorld.from_checkpoint(file_name)

    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    world = w
    file_name = world.checkpoint(tmp_path)
    w = DummyWorld.from_checkpoint(file_name)

    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    w.step()

    world = w
    file_name = world.checkpoint(tmp_path)
    w = DummyWorld.from_checkpoint(file_name)

    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    w.run()


def test_agent_checkpoint(tmp_path):

    a = DummyAgent(name="abcd")

    a.step_()

    file_name = a.checkpoint(tmp_path, info={"a": 3})

    info = SAOMechanism.checkpoint_info(file_name)
    assert isinstance(info["time"], str)
    assert info["step"] == 1
    assert info["type"].endswith("DummyAgent")
    assert info["id"] == a.id
    assert info["name"] == a.name == "abcd"
    assert info["a"] == 3

    b = a.from_checkpoint(file_name)

    assert a.id == b.id

    b.step_()


def test_agent_checkpoint_in_world(tmp_path):
    world = DummyWorld(n_steps=10)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    assert len(world.agents) == 2
    for a in world.agents.values():
        a.step_()
        file_name = a.checkpoint(tmp_path, info={"a": 3})

        info = SAOMechanism.checkpoint_info(file_name)
        assert isinstance(info["time"], str)
        assert info["step"] == 1
        assert info["type"].endswith("DummyAgent")
        assert info["id"] == a.id
        assert info["name"] == a.name
        assert info["a"] == 3

        b = a.from_checkpoint(file_name)

        assert a.id == b.id

        b.step_()


class MyMonitor(EventSink):
    def on_event(self, event: Event, sender: EventSource):
        assert event.type == "agent-joined"


def test_cannot_start_a_neg_with_no_outcomes():
    world = DummyWorld(n_steps=10)
    a, b = DummyAgent(name="a"), DummyAgent(name="b")
    world.join(a)
    world.join(b)
    assert not a.awi.request_negotiation_about(
        issues=[Issue((1, 0))], partners=[a.id, b.id], req_id="1234"
    )


def test_world_monitor():
    monitor = MyMonitor()
    world = DummyWorld(n_steps=10)
    world.register_listener("agent-joined", listener=monitor)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    assert len(world.agents) == 2
    world.run()


if __name__ == "__main__":
    pytest.main(args=[__file__])



import os
from typing import Union

import numpy as np
import pkg_resources
import pytest

from negmas import load_genius_domain_from_folder
from negmas.elicitation import (
    BalancedElicitor,
    DummyElicitor,
    EStrategy,
    FullKnowledgeElicitor,
    PandoraElicitor,
    SAOElicitingMechanism,
    User,
    next_query,
    possible_queries,
    OptimalIncrementalElicitor,
)
from negmas.helpers import Distribution as U, instantiate
from negmas.sao import AspirationNegotiator, LimitedOutcomesNegotiator, SAOMechanism
from negmas.utilities import IPUtilityFunction, MappingUtilityFunction, pareto_frontier

n_outcomes = 5
cost = 0.02
utility = 0.17682
accepted = [(0,), (2,)]
ufun = MappingUtilityFunction(
    dict(zip([(_,) for _ in range(n_outcomes)], [utility] * n_outcomes)),
    reserved_value=0.0,
)

import negmas.elicitation as elicitation


all_countable_queries_elicitor_types = [
    _
    for _ in elicitation.__all__
    if _.endswith("Elicitor") and not _.startswith("Base") and not "VOIOptimal" in _
]


@pytest.fixture
def neg() -> SAOMechanism:
    return SAOMechanism(outcomes=[(_,) for _ in range(n_outcomes)])


@pytest.fixture
def strategy(neg: SAOMechanism) -> EStrategy:
    s = EStrategy(strategy="exact")
    s.on_enter(ami=neg.ami)
    return s


@pytest.fixture
def user() -> User:
    return User(ufun=ufun, cost=cost)


@pytest.fixture
def true_utilities():
    return list(np.random.rand(n_outcomes).tolist())


@pytest.fixture
def master(true_utilities, strategy_name="titration-0.05"):
    user = User(
        ufun=MappingUtilityFunction(
            dict(zip([(_,) for _ in range(n_outcomes)], true_utilities)),
            reserved_value=0.0,
        ),
        cost=cost,
    )
    strategy = EStrategy(strategy=strategy_name)
    return user, strategy


class TestCountableOutcomesUser(object):
    def test_countable_outcmoes_user_initializable(self):
        user = User(ufun=ufun, cost=cost)
        assert user.total_cost == 0.0, "total cost is not initialized to zero"

    def test_countable_outcmoes_user_can_enter(self, user):
        true_utils = list(user.ufun.mapping.values())
        assert len(true_utils) == n_outcomes, "incorrect number of utilities"
        assert user.total_cost == 0.0
        assert user.ufun((0,)) == utility
        assert user.cost_of_asking() == cost

    def test_countable_outcmoes_user_can_elicit_exact(self, strategy, user):
        u, _ = strategy.apply(user=user, outcome=(0,))
        assert isinstance(u, float), "exact elicitation gets u a non float"
        assert u == utility, "exact elicitation gets u an incorrect value"

    def test_countable_outcmoes_user_can_elicit_bisection(self, neg, user):
        strategy = EStrategy(strategy="bisection", resolution=1e-4)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        while True:
            e = strategy.utility_estimate((0,))
            # assert user.total_cost <= total_cost + cost, 'incorrect total cost'
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                assert abs(u - utility) < 1e-2
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)

        assert elicited[0].loc == 0.0, "first loc is incorrect in elicitation"
        assert elicited[0].scale == 0.5, "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"
        for u, e in zip(elicited, estimated):
            assert u.scale == 0.5 * e.scale, "uncertainty is not decreasing as expected"
        for i in range(len(elicited) - 1):
            assert (
                elicited[i + 1].scale == 0.5 * elicited[i].scale
            ), "uncertainty is not decreasing"
        for i in range(len(estimated) - 1):
            assert (
                estimated[i + 1].scale == 0.5 * estimated[i].scale
            ), "uncertainty is not decreasing"

    def test_countable_outcmoes_user_can_elicit_titration_up(self, neg, user):
        step = 0.05
        strategy = EStrategy(strategy=f"titration+{step}", resolution=1e-4)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        while True:
            e = strategy.utility_estimate((0,))
            assert user.total_cost == total_cost, "incorrect total cost"
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                assert abs(u - utility) < step * 2
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)

        assert elicited[0].loc == step, "first loc is incorrect in elicitation"
        assert (
            elicited[0].scale == 1.0 - step
        ), "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"
        for u, e in zip(elicited[:-1], estimated[:-1]):
            assert (
                abs(u.scale - e.scale + step) < 1e-3
            ), "uncertainty is not decreasing as expected"
        for i in range(len(elicited) - 2):
            assert (
                abs(elicited[i + 1].scale - elicited[i].scale + step) < 1e-3
            ), "uncertainty is not decreasing"
        for i in range(len(estimated) - 2):
            assert (
                estimated[i + 1].scale - estimated[i].scale + step
            ) < 1e-3, "uncertainty is not decreasing"

    def test_countable_outcmoes_user_can_elicit_titration_up_no_step(self, neg, user):
        step = 0.01
        strategy = EStrategy(strategy=f"titration", resolution=step)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        for _ in range(int(0.5 + 1.0 / step) + 2):
            e = strategy.utility_estimate((0,))
            # assert user.total_cost <= total_cost + cost, 'incorrect total cost'
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                u = float(u)
                assert abs(u - utility) < 2 * step
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)
        else:
            # print(elicited)
            assert False, "did not end in expected time"

        assert elicited[0].loc == step, "first loc is incorrect in elicitation"
        assert (
            elicited[0].scale == 1.0 - step
        ), "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"
        for u, e in zip(elicited[:-1], estimated[:-1]):
            assert (
                abs(u.scale - e.scale + step) < 1e-3
            ), "uncertainty is not decreasing as expected"
        for i in range(len(elicited) - 2):
            assert (
                abs(elicited[i + 1].scale - elicited[i].scale + step) < 1e-3
            ), "uncertainty is not decreasing"
        for i in range(len(estimated) - 2):
            assert (
                estimated[i + 1].scale - estimated[i].scale + step
            ) < 1e-3, "uncertainty is not decreasing"

    def test_countable_outcmoes_user_can_elicit_titration_down(self, neg, user):
        step = -0.05
        strategy = EStrategy(strategy=f"titration{step}", resolution=1e-4)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        while True:
            e = strategy.utility_estimate((0,))
            # assert user.total_cost <= total_cost + cost, 'incorrect total cost'
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                assert abs(u - utility) < -step * 2
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)

        step = -step
        assert (
            elicited[0].loc + elicited[0].scale == 1.0 - step
        ), "first loc is incorrect in elicitation"
        assert (
            elicited[0].scale == 1.0 - step
        ), "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"

    def test_countable_outcmoes_user_can_elicit_dtitration_up(self, neg, user):
        step = 0.05
        strategy = EStrategy(strategy=f"dtitration+{step}", resolution=1e-4)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        while True:
            e = strategy.utility_estimate((0,))
            # assert user.total_cost <= total_cost + cost, 'incorrect total cost'
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                assert abs(u - utility) < 1e-2
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)
        assert elicited[0].loc == step, "first loc is incorrect in elicitation"
        assert (
            elicited[0].scale == 1.0 - step
        ), "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"

    def test_countable_outcmoes_user_can_elicit_dtitration_down(self, neg, user):
        step = -0.05
        strategy = EStrategy(strategy=f"dtitration{step}", resolution=1e-4)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        while True:
            e = strategy.utility_estimate((0,))
            # assert user.total_cost <= total_cost + cost, 'incorrect total cost'
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                assert abs(u - utility) < -step * 2
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)

        step = -step
        assert (
            elicited[0].loc + elicited[0].scale == 1.0 - step
        ), "first loc is incorrect in elicitation"
        assert (
            elicited[0].scale == 1.0 - step
        ), "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"

    def test_countable_outcmoes_user_stops_eliciting_at_cost(self, neg, user):
        strategy = EStrategy(strategy="bisection", resolution=1e-4, stop_at_cost=True)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        while True:
            e = strategy.utility_estimate((0,))
            # assert user.total_cost <= total_cost + cost, 'incorrect total cost'
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                assert abs(u - utility) < cost
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)

        assert elicited[0].loc == 0.0, "first loc is incorrect in elicitation"
        assert elicited[0].scale == 0.5, "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"
        for u, e in zip(elicited, estimated):
            assert u.scale == 0.5 * e.scale, "uncertainty is not decreasing as expected"
        for i in range(len(elicited) - 1):
            assert (
                elicited[i + 1].scale == 0.5 * elicited[i].scale
            ), "uncertainty is not decreasing"
        for i in range(len(estimated) - 1):
            assert (
                estimated[i + 1].scale == 0.5 * estimated[i].scale
            ), "uncertainty is not decreasing"
        assert estimated[-1].scale >= cost

    def test_countable_outcmoes_user_elicits_non_exact(self, neg, user):
        strategy = EStrategy(strategy="bisection", resolution=1e-2)
        strategy.on_enter(neg.ami)
        elicited = []
        estimated = []
        total_cost = 0.0
        while True:
            e = strategy.utility_estimate((0,))
            # assert user.total_cost <= total_cost + cost, 'incorrect total cost'
            u, reply = strategy.apply(user=user, outcome=(0,))
            if isinstance(u, float) or u.scale < cost:
                assert abs(u - utility) < 1e-2 and abs(u - utility) > 1e-3
                break
            total_cost += cost
            elicited.append(u)
            estimated.append(e)

        assert elicited[0].loc == 0.0, "first loc is incorrect in elicitation"
        assert elicited[0].scale == 0.5, "first scale is incorrect in elicitation"
        assert estimated[0].loc == 0.0, "first loc is incorrect in estimation"
        assert estimated[0].scale == 1.0, "first scale is incorrect in estimation"
        for u, e in zip(elicited, estimated):
            assert u.scale == 0.5 * e.scale, "uncertainty is not decreasing as expected"
        for i in range(len(elicited) - 1):
            assert (
                elicited[i + 1].scale == 0.5 * elicited[i].scale
            ), "uncertainty is not decreasing"
        for i in range(len(estimated) - 1):
            assert (
                estimated[i + 1].scale == 0.5 * estimated[i].scale
            ), "uncertainty is not decreasing"

    def test_countable_outcomes_user_can_return_all_queries(self, neg):
        for s in (
            "exact",
            "bisection",
            "titration+0.05",
            "titration-0.5",
            "dtitration+0.5",
            "dtitration-0.05",
            "pingpong0.5",
            "dpingpong0.5",
        ):
            user = User(ufun=ufun, cost=cost)
            strategy = EStrategy(strategy=s)
            strategy.on_enter(neg.ami)
            q = possible_queries(ami=neg.ami, strategy=strategy, user=user)
            assert (
                len(q) > 0 and s != "exact" or len(q) == 0 and s == "exact"
            ), "returns some queries"

    def test_countable_outcomes_user_can_return_next_queries(self, neg):
        for s in (
            "exact",
            "bisection",
            "titration+0.05",
            "titration-0.5",
            "dtitration+0.5",
            "dtitration-0.05",
        ):
            strategy = EStrategy(strategy=s)
            user = User(ufun=ufun, cost=cost)
            strategy.on_enter(neg.ami)
            q = next_query(strategy=strategy, user=user)
            # print(f'{strategy} Strategy:\n---------------')
            # pprint.pprint(q)
            assert len(q) > 0, "returns some queries"

    def test_elicit_until(self, neg):
        for s in ("bisection", "titration+0.05", "titration-0.05", "pingpong"):
            user = User(ufun=ufun, cost=cost)
            stretegy = EStrategy(strategy=s, resolution=1e-3)
            stretegy.on_enter(neg.ami)
            outcome, query, qcost = next_query(
                strategy=stretegy, outcome=(0,), user=user
            )[0]
            stretegy.until(
                user=user, outcome=(0,), dist=query.answers[1].constraint.marginal((0,))
            )


def u0(neg: SAOMechanism, reserved_value=0.0):
    return IPUtilityFunction(outcomes=neg.outcomes, reserved_value=reserved_value)


@pytest.fixture
def data_folder():
    return pkg_resources.resource_filename("negmas", resource_name="tests/data")


class TestCountableOutcomesElicitor(object):
    def test_dummy(self, master, true_utilities):
        user, strategy = master
        neg = SAOMechanism(outcomes=[(_,) for _ in range(n_outcomes)], n_steps=10)
        accepted = [(0,), (2,)]
        opponent = LimitedOutcomesNegotiator(
            acceptable_outcomes=accepted,
            acceptance_probabilities=[1.0] * len(accepted),
        )
        elicitor = DummyElicitor(user=user)
        neg.add(opponent)
        neg.add(elicitor, ufun=u0(neg))
        neg.run()
        # print(
        #     f'Got {elicitor.ufun(neg.agreement)} with elicitation cost {elicitor.elicitation_cost}'
        #     f' for {elicitor} using 0 elicited_queries')

        assert len(neg.history) > 0
        assert neg.agreement is None or neg.agreement in accepted
        # assert len(set([_[0] for _ in elicitor.offers])) > 1 or len(elicitor.offers) < 2
        assert elicitor.elicitation_cost == 0.0

    def test_full_knowledge(self, master):
        user, strategy = master
        neg = SAOMechanism(outcomes=[(_,) for _ in range(n_outcomes)], n_steps=10)
        opponent = LimitedOutcomesNegotiator(
            acceptable_outcomes=accepted,
            acceptance_probabilities=[1.0] * len(accepted),
        )
        elicitor = FullKnowledgeElicitor(user=user)
        neg.add(opponent)
        neg.add(elicitor, ufun=u0(neg))
        neg.run()
        # print(
        #     f'Got {elicitor.ufun(neg.agreement)} with elicitation cost {elicitor.elicitation_cost}'
        #     f' for {elicitor} using 0 elicited_queries')

        assert len(neg.history) > 0
        assert neg.agreement is None or neg.agreement in accepted
        # assert len(set([_[0] for _ in elicitor.offers])) > 1 or len(elicitor.offers) < 2
        assert elicitor.elicitation_cost == 0.0

    @pytest.mark.parametrize("elicitor", all_countable_queries_elicitor_types)
    def test_elicitor_runs(
        self, elicitor: Union[str, "BaseElicitor"], master, true_utilities, **kwargs
    ):
        neg = SAOMechanism(outcomes=n_outcomes, n_steps=10)
        user, strategy = master
        opponent = LimitedOutcomesNegotiator(
            acceptable_outcomes=accepted,
            acceptance_probabilities=[1.0] * len(accepted),
        )
        strategy.on_enter(ami=neg.ami)
        if isinstance(elicitor, str):
            elicitor = f"negmas.elicitation.{elicitor}"
            if "VOI" in elicitor:
                kwargs["dynamic_query_set"] = True
            elicitor = instantiate(elicitor, strategy=strategy, user=user, **kwargs)
        neg.add(opponent)
        neg.add(elicitor)
        assert elicitor.elicitation_cost == 0.0
        neg.run()
        queries = list(elicitor.user.elicited_queries())
        assert len(neg.history) > 0
        assert neg.agreement is None or neg.agreement in accepted
        assert (
            elicitor.elicitation_cost > 0.0
            or cost == 0.0
            or elicitor.strategy is None
            or neg.state.step < 2
        )
        if neg.agreement is not None:
            assert (
                elicitor.user_ufun(neg.agreement)
                == true_utilities[neg.agreement[0]] - elicitor.elicitation_cost
            )
        if hasattr(elicitor, "each_outcome_once") and elicitor.each_outcome_once:
            assert len(set([_[0] for _ in elicitor.offers])) == len(elicitor.offers)
        # print(
        #     f"Got {elicitor.ufun(neg.agreement)} with elicitation cost {elicitor.elicitation_cost} "
        #     f"for {elicitor} using {len(queries)} elicited_queries"
        # )

    # def test_pareto_frontier_in_mechanism(self, master, true_utilities):
    #     neg, elicitor, opponent = self._run_optimal_test(
    #         "FullElicitor", master, true_utilities
    #     )
    #     front, _ = neg.pareto_frontier()
    #     # print(front)
    #     assert len(front) > 0

    def test_pareto_frontier_2(self):
        n_outcomes = 10
        strategy = "titration-0.5"
        cost = 0.01
        reserved_value = 0.1
        outcomes = [(_,) for _ in range(n_outcomes)]
        accepted = [(2,), (3,), (4,), (5,)]
        elicitor_utilities = [
            0.5337723805661662,
            0.8532272031479199,
            0.4781281413197942,
            0.7242899747791032,
            0.3461879818432919,
            0.2608677043479706,
            0.9419131964655383,
            0.29368079952747694,
            0.6093201983562316,
            0.7066918086398718,
        ]
        # list(np.random.rand(n_outcomes).tolist())
        opponent_utilities = [
            1.0 if (_,) in accepted else 0.0 for _ in range(n_outcomes)
        ]
        frontier, frontier_locs = pareto_frontier(
            [
                MappingUtilityFunction(
                    lambda o: elicitor_utilities[o[0]],
                    reserved_value=reserved_value,
                    outcome_type=tuple,
                ),
                MappingUtilityFunction(
                    lambda o: opponent_utilities[o[0]],
                    reserved_value=reserved_value,
                    outcome_type=tuple,
                ),
            ],
            outcomes=outcomes,
            sort_by_welfare=True,
        )
        welfare = (
            np.asarray(elicitor_utilities) + np.asarray(opponent_utilities)
        ).tolist()
        # print(f'frontier: {frontier}\nmax. welfare: {max(welfare)} at outcome: ({welfare.index(max(welfare))},)')
        # print(f'frontier_locs: frontier_locs')
        neg = SAOMechanism(outcomes=n_outcomes, n_steps=10, outcome_type=tuple)
        opponent = LimitedOutcomesNegotiator(
            acceptable_outcomes=accepted,
            acceptance_probabilities=[1.0] * len(accepted),
        )
        eufun = MappingUtilityFunction(
            dict(zip(outcomes, elicitor_utilities)),
            reserved_value=reserved_value,
            outcome_type=tuple,
        )
        user = User(ufun=eufun, cost=cost)
        strategy = EStrategy(strategy=strategy)
        strategy.on_enter(ami=neg.ami)
        elicitor = FullKnowledgeElicitor(strategy=strategy, user=user)
        neg.add(opponent)
        neg.add(elicitor)
        neg.run()
        f2, f2_outcomes = neg.pareto_frontier(sort_by_welfare=True)
        assert len(frontier) == len(f2)
        assert all([_1 == _2] for _1, _2 in zip(frontier, f2))
        assert [_[0] for _ in f2_outcomes] == frontier_locs

    def test_loading_laptop(self, data_folder):
        domain, agents_info, issues = load_genius_domain_from_folder(
            os.path.join(data_folder, "Laptop"),
            force_single_issue=True,
            keep_issue_names=True,
            keep_value_names=True,
            agent_factories=lambda: AspirationNegotiator(),
        )
        # [domain.add(LimitedOutcomesNegotiator(outcomes=n_outcomes)
        #            , ufun=_['ufun']) for _ in agents_info]
        front, locs = domain.pareto_frontier(sort_by_welfare=True)
        assert front == [
            (0.7715533992081258, 0.8450562871935449),
            (0.5775524426410947, 1.0),
            (1.0, 0.5136317604069089),
            (0.8059990434329689, 0.6685754732133642),
        ]

    def test_loading_laptop_no_names(self, data_folder):
        domain, agents_info, issues = load_genius_domain_from_folder(
            os.path.join(data_folder, "Laptop"),
            force_single_issue=True,
            keep_issue_names=False,
            keep_value_names=False,
            agent_factories=lambda: AspirationNegotiator(),
        )
        # [domain.add(LimitedOutcomesNegotiator(outcomes=n_outcomes)
        #            , ufun=_['ufun']) for _ in agents_info]
        front, _ = domain.pareto_frontier(sort_by_welfare=True)
        assert front == [
            (0.7715533992081258, 0.8450562871935449),
            (0.5775524426410947, 1.0),
            (1.0, 0.5136317604069089),
            (0.8059990434329689, 0.6685754732133642),
        ]

    def test_elicitor_can_get_frontier(self, data_folder):
        domain, agents_info, issues = load_genius_domain_from_folder(
            os.path.join(data_folder, "Laptop"),
            force_single_issue=True,
            keep_issue_names=False,
            keep_value_names=False,
            normalize_utilities=True,
        )
        assert len(issues) == 1
        assert len(agents_info) == 2
        domain.add(LimitedOutcomesNegotiator(), ufun=agents_info[0]["ufun"])
        user = User(ufun=agents_info[0]["ufun"], cost=cost)
        strategy = EStrategy(strategy="titration-0.5")
        strategy.on_enter(ami=domain.ami)
        elicitor = FullKnowledgeElicitor(strategy=strategy, user=user)
        domain.add(elicitor)
        front, _ = domain.pareto_frontier()
        assert front == [(1.0, 1.0)]

    def test_elicitor_can_run_from_genius_domain(self, data_folder):
        domain, agents_info, issues = load_genius_domain_from_folder(
            os.path.join(data_folder, "Laptop"),
            force_single_issue=True,
            keep_issue_names=False,
            keep_value_names=False,
        )
        domain.add(LimitedOutcomesNegotiator(), ufun=agents_info[0]["ufun"])
        # domain.n_steps = 10
        user = User(ufun=agents_info[0]["ufun"], cost=0.2)
        strategy = EStrategy(strategy="titration-0.5")
        strategy.on_enter(ami=domain.ami)
        elicitor = PandoraElicitor(strategy=strategy, user=user)
        domain.add(elicitor)
        front, _ = domain.pareto_frontier()
        domain.run()
        assert len(domain.history) > 0

    def test_voi_baarslag_example(self):
        outcomes = [(0,), (1,)]
        alphas = [0.0, 1.0]
        opponent = LimitedOutcomesNegotiator(
            acceptable_outcomes=outcomes, acceptance_probabilities=alphas,
        )


def test_elicitation_run_with_no_conflict():
    n_outcomes = 50
    n_steps = 100
    config = SAOElicitingMechanism.generate_config(
        cost=0.05,
        n_outcomes=n_outcomes,
        conflict=0.0,
        winwin=1.0,
        n_steps=n_steps,
        own_reserved_value=0.1,
        opponent_type="tough",
        opponent_model_uncertainty=0.0,
        own_utility_uncertainty=0.0,
    )
    neg = SAOElicitingMechanism(**config, elicitor_type="full_knowledge")
    frontier, frontier_outcomes = neg.pareto_frontier(sort_by_welfare=True)
    assert len(frontier) > 0
    neg.run()
    assert neg.agreement is None or neg.agreement in frontier_outcomes


def test_alternating_offers_eliciting_mechanism():
    for strategy, dynamic in [("pingpong", False), ("bisection", True)]:
        config = SAOElicitingMechanism.generate_config(
            cost=0.001,
            n_outcomes=10,
            opponent_model_adaptive=False,
            opponent_type="limited_outcomes",
            conflict=1.0,
            n_steps=500,
            time_limit=100000.0,
            own_utility_uncertainty=0.1,
            own_reserved_value=0.1,
        )
        p = SAOElicitingMechanism(
            **config,
            elicitation_strategy=strategy,
            elicitor_type="balanced",
            dynamic_queries=dynamic,
        )
        p.run()
        assert len(p.history) > 0
        # pprint(p.elicitation_state)
        # print(p.agents[1].reserved_value)
        assert p.elicitation_state["elicitation_cost"] >= 0.0
        assert (
            p.elicitation_state["elicitor_utility"] >= p.negotiators[1].reserved_value
        )


def test_alternating_offers_eliciting_mechanism_voi():
    for strategy, dynamic in [(None, False), (None, True)]:
        config = SAOElicitingMechanism.generate_config(
            cost=0.001,
            n_outcomes=10,
            opponent_model_adaptive=False,
            opponent_type="limited_outcomes",
            conflict=1.0,
            n_steps=500,
            time_limit=100000.0,
            own_utility_uncertainty=0.1,
            own_reserved_value=0.1,
        )
        p = SAOElicitingMechanism(
            **config,
            elicitation_strategy=strategy,
            elicitor_type="voi",
            dynamic_queries=dynamic,
        )
        p.run()
        assert len(p.history) > 0
        # pprint(p.elicitation_state)
        # print(p.agents[1].reserved_value)
        assert p.elicitation_state["elicitation_cost"] >= 0.0
        assert (
            p.elicitation_state["elicitor_utility"] >= p.negotiators[1].reserved_value
        )


def test_alternating_offers_eliciting_mechanism_voi_optimal():
    config = SAOElicitingMechanism.generate_config(
        cost=0.001,
        n_outcomes=10,
        opponent_model_adaptive=False,
        opponent_type="limited_outcomes",
        conflict=1.0,
        n_steps=500,
        time_limit=100000.0,
        own_utility_uncertainty=0.5,
        own_reserved_value=0.1,
    )
    p = SAOElicitingMechanism(**config, elicitor_type="voi_optimal")
    p.run()
    assert len(p.history) > 0
    # pprint(p.elicitation_state)
    # print(p.agents[1].reserved_value)
    assert p.elicitation_state["elicitation_cost"] >= 0.0
    assert (
        p.elicitation_state["elicitor_utility"]
        + p.elicitation_state["elicitation_cost"]
        >= p.negotiators[1].reserved_value
    )
    assert (
        p.elicitation_state["total_voi"] is not None
        and p.elicitation_state["total_voi"] >= 0
    )


def test_alternating_offers_eliciting_mechanism_full_knowledge():
    config = SAOElicitingMechanism.generate_config(
        cost=0.001,
        n_outcomes=10,
        opponent_model_adaptive=False,
        opponent_type="limited_outcomes",
        conflict=1.0,
        n_steps=500,
        time_limit=100000.0,
        own_utility_uncertainty=0.1,
        own_reserved_value=0.1,
    )
    p = SAOElicitingMechanism(
        **config, elicitation_strategy="bisection", elicitor_type="full_knowledge"
    )
    p.run()
    assert len(p.history) > 0
    # pprint(p.elicitation_state)
    # print(p.agents[1].reserved_value)
    assert p.elicitation_state["elicitation_cost"] == 0.0
    assert p.elicitation_state["elicitor_utility"] >= p.negotiators[1].reserved_value


def test_a_small_elicitation_session():
    import os

    config = SAOElicitingMechanism.generate_config(cost=0.2, n_outcomes=5, n_steps=10)
    p = SAOElicitingMechanism(
        **config,
        history_file_name=f"{os.path.expanduser('~/logs/negmas/tmp/elicit.log')}'",
    )
    p.run()
    # pprint(p.history)


def test_a_typical_elicitation_session():
    import random
    import os

    n_outcomes = 10
    accepted_outcomes = [int(random.random() <= 0.5) for _ in range(n_outcomes)]
    config = SAOElicitingMechanism.generate_config(
        cost=0.2, n_outcomes=n_outcomes, n_steps=100, own_reserved_value=0.25
    )
    p = SAOElicitingMechanism(
        **config,
        history_file_name=f"{os.path.expanduser('~/logs/negmas/tmp/elicit.log')}'",
    )
    p.run()
    # pprint(p.elicitation_state)



import math
import sys
from functools import partial

import click
from PyQt5 import QtCore, QtGui, QtWidgets

MAX_ROWS = 21


try:
    import qdarkstyle

    _has_qdarkstyle = True
except ModuleNotFoundError:
    _has_qdarkstyle = False


_GTypeRole = QtCore.Qt.UserRole
_missing = object()


class GStyle(object):
    _base_style = """
        ._OptionLabel {
            font-size: 16px;
            font: bold;
            font-family: arial;
            }
        ._HelpLabel {
            font-family: arial;
            font-size: 14px;
            }
        ._InputComboBox{
            font-size: 16px;
            }
        ._InputLineEdit{
            font-size: 16px;
            }
        ._InputCheckBox{
            font-size: 16px;
            }
        ._InputSpinBox{
            font-size: 16px;
            }
        ._InputTabWidget{
            font: bold;
            font-size: 16px;
            }
        .GListView{
            font-size: 16px;
            }
        QToolTip{
            font-family: arial;
            }
        """

    def __init__(self, style=""):
        if not GStyle.check_style(style):
            self.text_color = "black"
            self.placehoder_color = "#898b8d"
            self.stylesheet = (
                GStyle._base_style
                + """
                    ._Spliter{
                        border: 1px inset gray;
                        }
                    """
            )
        elif style == "qdarkstyle":
            self.text_color = "#eff0f1"
            self.placehoder_color = "#898b8d"
            self.stylesheet = (
                qdarkstyle.load_stylesheet_pyqt5()
                + GStyle._base_style
                + """
                    .GListView{
                        padding: 5px;
                        }
                    ._Spliter{
                        border: 5px solid gray;
                        }
                    """
            )

    @staticmethod
    def check_style(style):
        if style == "qdarkstyle":
            return _has_qdarkstyle
        return False


_gstyle = GStyle()


class GListView(QtWidgets.QListView):
    def __init__(self, opt):
        super(GListView, self).__init__()
        self.nargs = opt.nargs
        self.model = GItemModel(
            opt.nargs, parent=self, opt_type=opt.type, default=opt.default
        )
        self.setModel(self.model)
        self.delegate = GEditDelegate(self)
        self.setItemDelegate(self.delegate)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        if self.nargs == -1:
            self.keyPressEvent = self.key_press
            self.setToolTip(
                "'a': add a new item blow the selected one\n"
                "'d': delete the selected item"
            )

    def key_press(self, e):
        if self.nargs == -1:
            if e.key() == QtCore.Qt.Key_A:
                if len(self.selectedIndexes()) == 0:
                    self.model.insertRow(0)
                else:
                    for i in self.selectedIndexes():
                        self.model.insertRow(i.row() + 1)
            if e.key() == QtCore.Qt.Key_D:
                si = self.selectedIndexes()
                for i in si:
                    self.model.removeRow(i.row())
        super(GListView, self).keyPressEvent(e)


class GItemModel(QtGui.QStandardItemModel):
    def __init__(self, n, parent=None, opt_type=click.STRING, default=None):
        super(QtGui.QStandardItemModel, self).__init__(0, 1, parent)
        self.type = opt_type
        for row in range(n):
            if hasattr(default, "__len__"):
                self.insertRow(row, default[row])
            else:
                self.insertRow(row, default)

    def insertRow(self, idx, val=""):
        super(GItemModel, self).insertRow(idx)

        index = self.index(idx, 0, QtCore.QModelIndex())
        if val is None or val == "":
            self.setData(
                index,
                QtGui.QBrush(QtGui.QColor(_gstyle.placehoder_color)),
                role=QtCore.Qt.ForegroundRole,
            )
        else:
            self.setData(index, val)

    def data(self, index, role=QtCore.Qt.DisplayRole):

        if role == QtCore.Qt.DisplayRole:
            dstr = QtGui.QStandardItemModel.data(self, index, role)
            if dstr == "" or dstr is None:
                if isinstance(self.type, click.types.Tuple):
                    row = index.row()
                    if 0 <= row < len(self.type.types):
                        tp = self.type.types[row]
                        dstr = tp.name
                else:
                    dstr = self.type.name
                return dstr

        if role == _GTypeRole:
            tp = click.STRING
            if isinstance(self.type, click.types.Tuple):
                row = index.row()
                if 0 <= row < len(self.type.types):
                    tp = self.type.types[row]
            elif isinstance(self.type, click.types.ParamType):
                tp = self.type
            return tp

        return QtGui.QStandardItemModel.data(self, index, role)


class GEditDelegate(QtWidgets.QStyledItemDelegate):
    def createEditor(self, parent, option, index):
        tp = index.data(role=_GTypeRole)
        if isinstance(tp, click.Path):
            led = GLineEdit_path.from_option(tp, parent)
        else:
            led = QtWidgets.QLineEdit(parent)
        led.setPlaceholderText(tp.name)
        led.setValidator(select_type_validator(tp))
        return led

    def setEditorData(self, editor, index):
        item_var = index.data(role=QtCore.Qt.EditRole)
        if item_var is not None:
            editor.setText(str(item_var))

    def setModelData(self, editor, model, index):
        data_str = editor.text()
        if data_str == "" or data_str is None:
            model.setData(
                index,
                QtGui.QBrush(QtGui.QColor(_gstyle.placehoder_color)),
                role=QtCore.Qt.ForegroundRole,
            )
        else:
            model.setData(
                index,
                QtGui.QBrush(QtGui.QColor(_gstyle.text_color)),
                role=QtCore.Qt.ForegroundRole,
            )
        QtWidgets.QStyledItemDelegate.setModelData(self, editor, model, index)


def generate_label(opt):
    show_name = getattr(opt, "show_name", _missing)
    show_name = opt.name if show_name is _missing else show_name
    param = _OptionLabel(show_name)
    param.setToolTip(getattr(opt, "help", None))
    return param


class GStringLineEditor(click.types.StringParamType):
    def to_widget(self, opt, validator=None):
        value = _InputLineEdit()
        value.setPlaceholderText(self.name)
        if opt.default:
            value.setText(str(opt.default))
        if getattr(opt, "hide_input", False):
            value.setEchoMode(QtWidgets.QLineEdit.Password)
        value.setValidator(validator)

        def to_command():
            return [opt.opts[0], value.text()]

        return [generate_label(opt), value], to_command


class GIntLineEditor(GStringLineEditor):
    def to_widget(self, opt):
        return GStringLineEditor.to_widget(self, opt, validator=QtGui.QIntValidator())


class GFloatLineEditor(GStringLineEditor):
    def to_widget(self, opt):
        return GStringLineEditor.to_widget(
            self, opt, validator=QtGui.QDoubleValidator()
        )


class GFileDialog(QtWidgets.QFileDialog):
    def __init__(self, *args, exists=False, file_okay=True, dir_okay=True, **kwargs):
        super(GFileDialog, self).__init__(*args, **kwargs)
        self.setOption(QtWidgets.QFileDialog.DontUseNativeDialog, True)
        self.setLabelText(QtWidgets.QFileDialog.Accept, "Select")
        if (exists, file_okay, dir_okay) == (True, True, False):
            self.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        elif (exists, file_okay, dir_okay) == (False, True, False):
            self.setFileMode(QtWidgets.QFileDialog.AnyFile)
        elif (exists, file_okay, dir_okay) == (True, False, True):
            self.setFileMode(QtWidgets.QFileDialog.Directory)
        elif (exists, file_okay, dir_okay) == (False, False, True):
            self.setFileMode(QtWidgets.QFileDialog.Directory)
        elif exists == True:
            self.setFileMode(QtWidgets.QFileDialog.ExistingFile)
            self.accept = self.accept_all
        elif exists == False:
            self.setFileMode(QtWidgets.QFileDialog.AnyFile)
            self.accept = self.accept_all

    def accept_all(self):
        super(GFileDialog, self).done(QtWidgets.QFileDialog.Accepted)


class GLineEdit_path(QtWidgets.QLineEdit):
    def __init__(self, parent=None, exists=False, file_okay=True, dir_okay=True):
        super(GLineEdit_path, self).__init__(parent)
        self.action = self.addAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon),
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self.fdlg = GFileDialog(
            self,
            "Select File Dialog",
            "./",
            "*",
            exists=exists,
            file_okay=file_okay,
            dir_okay=dir_okay,
        )
        self.action.triggered.connect(self.run_dialog)

    def run_dialog(self):
        if self.fdlg.exec() == QtWidgets.QFileDialog.Accepted:
            self.setText(self.fdlg.selectedFiles()[0])

    @staticmethod
    def from_option(opt, parent=None):
        return GLineEdit_path(
            parent=parent,
            exists=opt.exists,
            file_okay=opt.file_okay,
            dir_okay=opt.dir_okay,
        )


class GPathGLindEidt_path(click.types.Path):
    def to_widget(self, opt):
        value = GLineEdit_path(
            exists=self.exists, file_okay=self.file_okay, dir_okay=self.dir_okay
        )
        value.setPlaceholderText(self.name)
        if opt.default:
            value.setText(str(opt.default))

        def to_command():
            return [opt.opts[0], value.text()]

        return [generate_label(opt), value], to_command


class _GLabeledSlider(QtWidgets.QSlider):
    def __init__(self, min, max, val):
        super(_GLabeledSlider, self).__init__(QtCore.Qt.Horizontal)
        self.min, self.max = min, max

        self.setMinimum(min)
        self.setMaximum(max)
        self.setValue(val)

        self.label = self.__init_label()

    def __init_label(self):
        l = max(
            [
                math.ceil(math.log10(abs(x))) if x != 0 else 1
                for x in [self.min, self.max]
            ]
        )
        l += 1
        return QtWidgets.QLabel("0" * l)


def argument_command(to_command):
    def tc():
        a = to_command()
        return a[1:]

    return tc


class GSlider(QtWidgets.QHBoxLayout):
    def __init__(self, min=0, max=10, default=None, *args, **kwargs):
        super(QtWidgets.QHBoxLayout, self).__init__()

        self.min, self.max, self.default = min, max, default
        self.label = self.__init_label()
        self.slider = self.__init_slider()

        self.label.setText(str(self.default))

        self.addWidget(self.slider)
        self.addWidget(self.label)

    def value(self):
        return self.slider.value()

    def __init_slider(self):
        slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        slider.setMinimum(self.min)
        slider.setMaximum(self.max)
        default_val = (self.min + self.max) // 2
        if isinstance(self.default, int):
            if self.min <= self.default <= self.max:
                default_val = self.default
        self.default = default_val
        slider.setValue(default_val)
        slider.valueChanged.connect(lambda x: self.label.setText(str(x)))
        return slider

    def __init_label(self):
        l = max(
            [
                math.ceil(math.log10(abs(x))) if x != 0 else 1
                for x in [self.min, self.max]
            ]
        )
        l += 1
        return QtWidgets.QLabel("0" * l)


class GIntRangeGSlider(click.types.IntRange):
    def to_widget(self, opt):
        value = GSlider(min=self.min, max=self.max, default=opt.default)

        def to_command():
            return [opt.opts[0], str(value.value())]

        return [generate_label(opt), value], to_command


class GIntRangeSlider(click.types.IntRange):
    def to_widget(self, opt):
        value = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        value.setMinimum(self.min)
        value.setMaximum(self.max)

        default_val = (self.min + self.max) // 2
        if isinstance(opt.default, int):
            if self.min <= opt.default <= self.max:
                default_val = opt.default
        value.setValue(default_val)

        def to_command():
            return [opt.opts[0], str(value.value())]

        return [generate_label(opt), value], to_command


class GIntRangeLineEditor(click.types.IntRange):
    def to_widget(self, opt):
        value = QtWidgets.QLineEdit()
        # TODO: set validator

        def to_command():
            return [opt.opts[0], value.text()]

        return [generate_label(opt), value], to_command


def bool_flag_option(opt):
    checkbox = _InputCheckBox(opt.name)
    if opt.default:
        checkbox.setCheckState(2)
    # set tip
    checkbox.setToolTip(opt.help)

    def to_command():
        if checkbox.checkState():
            return [opt.opts[0]]
        else:
            return opt.secondary_opts

    return [checkbox], to_command


class GChoiceComboBox(click.types.Choice):
    def to_widget(self, opt):
        cb = _InputComboBox()
        cb.addItems(self.choices)

        def to_command():
            return [opt.opts[0], cb.currentText()]

        return [generate_label(opt), cb], to_command


def count_option(opt):
    sb = _InputSpinBox()

    def to_command():
        return [opt.opts[0]] * int(sb.text())

    return [generate_label(opt), sb], to_command


class GTupleGListView(click.Tuple):
    def to_widget(self, opt):
        view = GListView(opt)

        def to_command():
            _ = [opt.opts[0]]
            for idx in range(view.model.rowCount()):
                _.append(view.model.item(idx).text())
            return _

        return [generate_label(opt), view], to_command


def multi_text_arguement(opt):
    value = GListView(opt)

    def to_command():
        _ = []
        for idx in range(value.model.rowCount()):
            _.append(value.model.item(idx).text())
        # if opt.required and value.model.rowCount() == 0:
        # raise click.exceptions.BadParameter("Required")
        # print(opt.__dict__)
        return _

    # return [QtWidgets.QLabel(opt.name), value], to_command
    return [_OptionLabel(opt.name), value], to_command


def select_type_validator(tp: click.types.ParamType) -> QtGui.QValidator:
    """ select the right validator for `tp`"""

    if isinstance(tp, click.types.IntParamType):
        return QtGui.QIntValidator()
    elif isinstance(tp, click.types.FloatParamType):
        return QtGui.QDoubleValidator()

    return None


def select_opt_validator(opt):
    """ select the right validator for `opt`"""

    return select_type_validator(opt.type)


def opt_to_widget(opt):
    if opt.nargs > 1:
        return GTupleGListView.to_widget(opt.type, opt)
    elif getattr(opt, "is_bool_flag", False):
        return bool_flag_option(opt)
    elif getattr(opt, "count", False):
        return count_option(opt)
    elif isinstance(opt.type, click.types.Choice):
        return GChoiceComboBox.to_widget(opt.type, opt)
    elif isinstance(opt.type, click.types.Path):
        return GPathGLindEidt_path.to_widget(opt.type, opt)
    elif isinstance(opt.type, click.types.IntRange):
        return GIntRangeGSlider.to_widget(opt.type, opt)
    elif isinstance(opt.type, click.types.IntParamType):
        return GIntLineEditor.to_widget(opt.type, opt)
    elif isinstance(opt.type, click.types.FloatParamType):
        return GFloatLineEditor.to_widget(opt.type, opt)
    else:
        return GStringLineEditor.to_widget(opt.type, opt)


def _to_widget(opt):
    # customed widget

    if isinstance(opt.type, click.types.FuncParamType):
        if hasattr(opt.type.func, "to_widget"):
            return opt.type.func.to_widget()
    elif hasattr(opt.type, "to_widget"):
        return opt.type.to_widget()

    if isinstance(opt, click.core.Argument):
        if opt.nargs == 1:
            w, tc = opt_to_widget(opt)

            return w, argument_command(tc)
        elif opt.nargs > 1 or opt.nargs == -1:
            return multi_text_arguement(opt)
    else:
        return opt_to_widget(opt)


def layout_append_opts(layout, opts):
    params_func = []
    widgets = []
    i = 0

    for i, para in enumerate(opts):
        widget, value_func = _to_widget(para)
        widgets.append(widget)
        params_func.append(value_func)

        for idx, w in enumerate(widget):

            if isinstance(w, QtWidgets.QLayout):
                layout.addLayout(w, i, idx)
            else:
                layout.addWidget(w, i, idx)

    return layout, params_func, widgets


def generate_sysargv(cmd_list):
    argv_list = []

    for name, func_list in cmd_list:
        argv_list.append(name)

        for value_func in func_list:
            argv_list += value_func()

    return argv_list


class _Spliter(QtWidgets.QFrame):
    def __init__(self, parent=None):
        super(_Spliter, self).__init__(parent=parent)
        self.setFrameShape(QtWidgets.QFrame.HLine)


class _InputComboBox(QtWidgets.QComboBox):
    pass


class _InputTabWidget(QtWidgets.QTabWidget):
    pass


class _HelpLabel(QtWidgets.QLabel):
    pass


class _OptionLabel(QtWidgets.QLabel):
    pass


class _InputLineEdit(QtWidgets.QLineEdit):
    pass


class _InputCheckBox(QtWidgets.QCheckBox):
    pass


class _InputSpinBox(QtWidgets.QSpinBox):
    pass


class CommandLayout(QtWidgets.QGridLayout):
    def __init__(self, func, run_exit, parent_layout=None):
        super(CommandLayout, self).__init__()
        self.parent_layout = parent_layout
        self.func = func
        self.run_exit = run_exit

        if func.help:
            label = _HelpLabel(func.help)
            label.setWordWrap(True)
            self.addWidget(label, 0, 0, 1, 2)
            frame = _Spliter()
            self.addWidget(frame, 1, 0, 1, 4)
        self.params_func, self.widgets = self.append_opts(self.func.params)

    def add_sysargv(self):
        if hasattr(self.parent_layout, "add_sysargv"):
            self.parent_layout.add_sysargv()
        sys.argv += generate_sysargv([(self.func.name, self.params_func)])

    def append_opts(self, opts):
        params_func = []
        widgets = []

        def _nwidgets(x):
            return -len(_to_widget(x)[0])

        opts = sorted(opts, key=_nwidgets)
        row, col = 0, 0
        for para in opts:
            widget, value_func = _to_widget(para)
            widgets.append(widget)
            params_func.append(value_func)

            for idx, w in enumerate(widget):
                if isinstance(w, QtWidgets.QLayout):
                    self.addLayout(w, row + 2, 2 * col + idx)
                else:
                    self.addWidget(w, row + 2, 2 * col + idx)
            self.setRowStretch(row, 5)
            row += 1
            if row >= MAX_ROWS - 1:
                row = 0
                col += 1

        return params_func, widgets

    def generate_cmd_button(self, label, cmd_slot, tooltip=""):
        button = QtWidgets.QPushButton(label)
        button.setToolTip(tooltip)
        button.clicked.connect(self.clean_sysargv)
        button.clicked.connect(self.add_sysargv)
        button.clicked.connect(cmd_slot)

        return button

    def add_cmd_button(self, label, cmd_slot, pos=None):
        run_button = self.generate_cmd_button(label, cmd_slot)

        if pos is None:
            # pos = self.rowCount() + 1, 0
            pos = 0, 2
        self.addWidget(run_button, pos[0], pos[1])

    def add_cmd_buttons(self, args):
        # row = self.rowCount() + 1
        row = 0
        cmd_layout = QtWidgets.QGridLayout()
        cmd_layout.setHorizontalSpacing(20)

        for col, arg in enumerate(args):
            button = self.generate_cmd_button(**arg)
            cmd_layout.addWidget(button, col, 0)
        self.addLayout(cmd_layout, row, 2, 1, 2)

    @QtCore.pyqtSlot()
    def clean_sysargv(self):
        sys.argv = []


class RunCommand(QtCore.QRunnable):
    def __init__(self, func, run_exit):
        super(RunCommand, self).__init__()
        self.func = func
        self.run_exit = run_exit

    @QtCore.pyqtSlot()
    def run(self):
        print(sys.argv)
        try:
            self.func(standalone_mode=self.run_exit)
        except click.exceptions.BadParameter as bpe:
            # warning message
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText(bpe.format_message())
            msg.exec_()
        except Exception as bpe:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText(repr(bpe))
            msg.exec_()
        # if self.outputEdit is not None:
        # self.outputEdit.show()


class GCommand(click.Command):
    def __init__(self, new_thread=True, *arg, **args):
        super(GCommand, self).__init__(*arg, **args)
        self.new_thread = new_thread


class GOption(click.Option):
    def __init__(self, *arg, show_name=_missing, **args):
        super(GOption, self).__init__(*arg, **args)
        self.show_name = show_name


# def normalOutputWritten(t):
# """Append text to the QTextEdit."""
# Maybe QTextEdit.append() works as well, but this is how I do it:
# cursor = text.textCursor()
# cursor.movePosition(QtGui.QTextCursor.End)
# cursor.insertText(t)
# text.setTextCursor(cursor)
# text.ensureCursorVisible()


class GuiStream(QtCore.QObject):
    textWritten = QtCore.pyqtSignal(str)

    def flush(self):
        pass

    def write(self, text):
        self.textWritten.emit(str(text))


class OutputEdit(QtWidgets.QTextEdit):
    def print(self, text):
        cursor = self.textCursor()
        cursor.movePosition(QtGui.QTextCursor.End)
        cursor.insertText(text)
        self.setTextCursor(cursor)
        self.ensureCursorVisible()


class App(QtWidgets.QWidget):
    def __init__(
        self,
        func,
        run_exit,
        new_thread,
        output="gui",
        left=10,
        top=10,
        width=800,
        height=280,
    ):
        """
        Parameters
        ----------
        output : str
            'gui': [default] redirect screen output to the gui
            'term': do nothing
        """
        super().__init__()
        self.new_thread = new_thread
        self.title = func.name
        self.func = func
        self.initUI(run_exit, QtCore.QRect(left, top, width, height))
        self.threadpool = QtCore.QThreadPool()
        self.outputEdit = self.initOutput(output)

    def initOutput(self, output):
        if output == "gui":
            sys.stdout = GuiStream()
            sys.stderr = sys.stdout
            text = OutputEdit()
            text.setReadOnly(True)
            sys.stdout.textWritten.connect(text.print)
            sys.stdout.textWritten.connect(text.show)
            # sys.stdout.textWritten.connect(text.append)
            # text.show()

            return text
        else:
            return None

    def initCommandUI(self, func, run_exit, parent_layout=None):
        opt_set = CommandLayout(func, run_exit, parent_layout=parent_layout)

        if isinstance(func, click.MultiCommand):
            tabs = _InputTabWidget()

            for cmd, f in func.commands.items():
                sub_opt_set = self.initCommandUI(f, run_exit, parent_layout=opt_set)
                tab = QtWidgets.QWidget()
                tab.setLayout(sub_opt_set)
                tabs.addTab(tab, cmd)
            opt_set.addWidget(tabs, opt_set.rowCount(), 0, 1, 4)
            # return opt_set
        elif isinstance(func, click.Command):
            new_thread = getattr(func, "new_thread", self.new_thread)
            opt_set.add_cmd_buttons(
                args=[
                    {
                        "label": "&Run",
                        "cmd_slot": partial(self.run_cmd, new_thread=new_thread),
                        "tooltip": "run command",
                    },
                    {
                        "label": "&Copy",
                        "cmd_slot": self.copy_cmd,
                        "tooltip": "copy command to clipboard",
                    },
                ]
            )

        return opt_set

    def initUI(self, run_exit, geometry):
        self.run_exit = run_exit
        self.setWindowTitle(self.title)
        # self.setGeometry(self.left, self.top, self.width, self.height)
        self.setGeometry(geometry)
        self.opt_set = self.initCommandUI(self.func, run_exit,)
        self.setLayout(self.opt_set)
        self.show()

    @QtCore.pyqtSlot()
    def copy_cmd(self):
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cmd_text = " ".join(sys.argv)
        cb.setText(cmd_text, mode=cb.Clipboard)

        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setText(f"copy '{cmd_text}' to clipboard")
        msg.exec_()

    def run_cmd(self, new_thread):
        runcmd = RunCommand(self.func, self.run_exit)

        if new_thread:
            self.threadpool.start(runcmd)
        else:
            runcmd.run()


def gui_it(click_func, style="qdarkstyle", **argvs) -> None:
    """
    Parameters
    ----------
    click_func
    `new_thread` is used for qt-based func, like matplotlib
    """
    global _gstyle
    _gstyle = GStyle(style)
    app = QtWidgets.QApplication(sys.argv)
    app.setStyleSheet(_gstyle.stylesheet)

    # set the default value for argvs
    argvs["run_exit"] = argvs.get("run_exit", False)
    argvs["new_thread"] = argvs.get("new_thread", False)

    _ = App(click_func, **argvs)
    sys.exit(app.exec_())


def gui_option(f: click.core.BaseCommand) -> click.core.BaseCommand:
    """decorator for adding '--gui' option to command"""
    # TODO: add run_exit, new_thread
    def run_gui_it(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        f.params = [p for p in f.params if not p.name == "gui"]
        gui_it(f)
        ctx.exit()

    return click.option(
        "--gui",
        is_flag=True,
        callback=run_gui_it,
        help="run with gui",
        expose_value=False,
        is_eager=False,
    )(f)



"""
Implements negotiators for the SAO mechanism.
"""
import math
import random
import warnings
from abc import abstractmethod
from typing import Callable, List, Optional, Tuple, Type, Union

import numpy as np

from .common import SAOResponse
from .components import (
    LimitedOutcomesAcceptorMixin,
    LimitedOutcomesMixin,
    RandomProposalMixin,
    RandomResponseMixin,
)
from ..common import (
    AgentMechanismInterface,
    MechanismState,
    _ShadowAgentMechanismInterface,
)
from ..events import Notification
from ..java import (
    JavaCallerMixin,
    JNegmasGateway,
    from_java,
    java_link,
    python_identifier,
    to_dict,
    to_java,
)
from ..negotiators import AspirationMixin, Controller, Negotiator
from ..outcomes import Issue, Outcome, ResponseType, outcome_as_dict, outcome_as_tuple
from ..utilities import (
    JavaUtilityFunction,
    LinearUtilityFunction,
    MappingUtilityFunction,
    UtilityFunction,
    UtilityValue,
    outcome_with_utility,
    utility_range,
)

__all__ = [
    "SAONegotiator",
    "RandomNegotiator",
    "LimitedOutcomesNegotiator",
    "LimitedOutcomesAcceptor",
    "AspirationNegotiator",
    "ToughNegotiator",
    "OnlyBestNegotiator",
    "NaiveTitForTatNegotiator",
    "SimpleTitForTatNegotiator",
    "NiceNegotiator",
    "JavaSAONegotiator",
    "PassThroughSAONegotiator",
    "_ShadowSAONegotiator",
]


class SAONegotiator(Negotiator):
    """
    Base class for all SAO negotiators.

    Args:
         name: Negotiator name
         parent: Parent controller if any
         ufun: The ufun of the negotiator
         assume_normalized: If true, the negotiator can assume that the ufun is normalized between zreo and one.
         rational_proposal: If `True`, the negotiator will never propose something with a utility value less than its
                            reserved value. If `propose` returned such an outcome, a NO_OFFER will be returned instead.
         owner: The `Agent` that owns the negotiator.

    Remarks:
        - The only method that **must** be implemented by any SAONegotiator is `propose`.
        - The default `respond` method, accepts offers with a utility value no less than whatever `propose` returns
          with the same mechanism state.

    """

    def __init__(
        self,
        assume_normalized=True,
        ufun: Optional[UtilityFunction] = None,
        name: Optional[str] = None,
        rational_proposal=True,
        parent: Controller = None,
        owner: "Agent" = None,
    ):
        super().__init__(name=name, ufun=ufun, parent=parent, owner=owner)
        self.assume_normalized = assume_normalized
        self.__end_negotiation = False
        self.my_last_proposal: Optional["Outcome"] = None
        self.my_last_proposal_utility: float = None
        self.rational_proposal = rational_proposal
        self.add_capabilities({"respond": True, "propose": True, "max-proposals": 1})

    def on_notification(self, notification: Notification, notifier: str):
        """
        Called whenever a notification is received

        Args:
            notification: The notification
            notifier: The notifier entity

        Remarks:
            - The default implementation only responds to end_negotiation by ending the negotiation
        """
        if notification.type == "end_negotiation":
            self.__end_negotiation = True

    def propose_(self, state: MechanismState) -> Optional["Outcome"]:
        """
        The method directly called by the mechanism (through `counter` ) to ask for a proposal

        Args:
            state: The mechanism state

        Returns:
            An outcome to offer or None to refuse to offer

        Remarks:
            - Depending on the `SAOMechanism` settings, refusing to offer may be interpreted as ending the negotiation
            - The negotiator will only receive this call if it has the 'propose' capability.
            - Rational proposal is implemented in this method.

        """
        if not self._capabilities["propose"] or self.__end_negotiation:
            return None
        if self._ufun_modified:
            self.on_ufun_changed()
        proposal = self.propose(state=state)

        # never return a proposal that is less than the reserved value
        if self.rational_proposal:
            utility = None
            if proposal is not None and self._utility_function is not None:
                utility = self._utility_function(proposal)
                if utility is not None and utility < self.reserved_value:
                    return None

            if utility is not None:
                self.my_last_proposal = proposal
                self.my_last_proposal_utility = utility

        return proposal

    def respond_(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        """The method to be called directly by the mechanism (through `counter` ) to respond to an offer.

        Args:
            state: a `MechanismState` giving current state of the negotiation.
            offer: the offer being responded to.

        Returns:
            ResponseType: The response to the offer. Possible values are:

                - NO_RESPONSE: refuse to offer. Depending on the mechanism settings this may be interpreted as ending
                               the negotiation.
                - ACCEPT_OFFER: Accepting the offer.
                - REJECT_OFFER: Rejecting the offer. The negotiator will be given the chance to counter this
                                offer through a call of `propose_` later if this was not the last offer to be evaluated
                                by the mechanism.
                - END_NEGOTIATION: End the negotiation
                - WAIT: Instructs the mechanism to wait for this negotiator more. It may lead to cycles so use with care.

        Remarks:
            - The default implementation never ends the negotiation except if an earler end_negotiation notification is
              sent to the negotiator
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).

        """
        if self.__end_negotiation:
            return ResponseType.END_NEGOTIATION
        if self._ufun_modified:
            self.on_ufun_changed()
        return self.respond(state=state, offer=offer)

    def counter(
        self, state: MechanismState, offer: Optional["Outcome"]
    ) -> "SAOResponse":
        """
        Called by the mechanism to counter the offer. It just calls `respond_` and `propose_` as needed.

        Args:
            state: `MechanismState` giving current state of the negotiation.
            offer: The offer to be countered. None means no offer and the agent is requested to propose an offer

        Returns:
            Tuple[ResponseType, Outcome]: The response to the given offer with a counter offer if the response is REJECT

        """
        if self.__end_negotiation:
            return SAOResponse(ResponseType.END_NEGOTIATION, None)
        if offer is None:
            return SAOResponse(ResponseType.REJECT_OFFER, self.propose_(state=state))
        response = self.respond_(state=state, offer=offer)
        if response != ResponseType.REJECT_OFFER:
            return SAOResponse(response, None)
        return SAOResponse(response, self.propose_(state=state))

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        """Called to respond to an offer. This is the method that should be overriden to provide an acceptance strategy.

        Args:
            state: a `MechanismState` giving current state of the negotiation.
            offer: offer being tested

        Returns:
            ResponseType: The response to the offer

        Remarks:
            - The default implementation never ends the negotiation
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).

        """
        if self._utility_function is None:
            return ResponseType.REJECT_OFFER
        if offer is None:
            return ResponseType.REJECT_OFFER
        if self._utility_function(offer) < self.reserved_value:
            return ResponseType.REJECT_OFFER
        utility = None
        if self.my_last_proposal_utility is not None:
            utility = self.my_last_proposal_utility
        if utility is None:
            myoffer = self.propose_(state=state)
            if myoffer is None:
                return ResponseType.NO_RESPONSE
            utility = self._utility_function(myoffer)
        if utility is not None and self._utility_function(offer) >= utility:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def on_partner_proposal(
        self, state: MechanismState, partner_id: str, offer: "Outcome"
    ) -> None:
        """
        A callback called by the mechanism when a partner proposes something

        Args:
            state: `MechanismState` giving the state of the negotiation when the offer was porposed.
            partner_id: The ID of the agent who proposed
            offer: The proposal.

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_refused_to_propose(
        self, state: MechanismState, agent_id: str
    ) -> None:
        """
        A callback called by the mechanism when a partner refuses to propose

        Args:
            state: `MechanismState` giving the state of the negotiation when the partner refused to offer.
            agent_id: The ID of the agent who refused to propose

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_response(
        self,
        state: MechanismState,
        partner_id: str,
        outcome: "Outcome",
        response: "ResponseType",
    ) -> None:
        """
        A callback called by the mechanism when a partner responds to some offer

        Args:
            state: `MechanismState` giving the state of the negotiation when the partner responded.
            partner_id: The ID of the agent who responded
            outcome: The proposal being responded to.
            response: The response

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    @abstractmethod
    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        """Propose an offer or None to refuse.

        Args:
            state: `MechanismState` giving current state of the negotiation.

        Returns:
            The outcome being proposed or None to refuse to propose

        Remarks:
            - This function guarantees that no agents can propose something with a utility value

        """

    class Java:
        implements = ["jnegmas.sao.SAONegotiator"]


class RandomNegotiator(RandomResponseMixin, RandomProposalMixin, SAONegotiator):
    """
    A negotiation agent that responds randomly in a single negotiation.

    Args:
        p_acceptance: Probability of accepting an offer
        p_rejection:  Probability of rejecting an offer
        p_ending: Probability of ending the negotiation at any round
        can_propose: Whether the agent can propose or not
        **kwargs: Passed to the SAONegotiator

    Remarks:
        - If p_acceptance + p_rejection + p_ending < 1, the rest is the probability of no-response.
        - This negotiator ignores the `rational_proposal` parameter.
    """

    def propose_(self, state: MechanismState) -> Optional["Outcome"]:
        return RandomProposalMixin.propose(self, state)

    def respond_(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        return RandomResponseMixin.respond(self, state, offer)

    def __init__(
        self,
        p_acceptance=0.15,
        p_rejection=0.75,
        p_ending=0.1,
        can_propose=True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.init_random_response(
            p_acceptance=p_acceptance, p_rejection=p_rejection, p_ending=p_ending
        )
        self.init_random_proposal()
        self.capabilities["propose"] = can_propose

    def join(
        self,
        ami: AgentMechanismInterface,
        state: MechanismState,
        *,
        ufun: Optional["UtilityFunction"] = None,
        role: str = "agent",
    ) -> bool:
        """
        Will create a random utility function to be used by the negotiator.

        Args:
            ami: The AMI
            state: The current mechanism state
            ufun: IGNORED.
            role: IGNORED.
        """
        result = super().join(ami, state, ufun=ufun, role=role)
        if not result:
            return False
        if ami.issues is not None and any(_.is_continuous() for _ in ami.issues):
            self._utility_function = LinearUtilityFunction(
                weights=np.random.random(len(ami.issues))
            )
        else:
            outcomes = [outcome_as_tuple(_) for _ in ami.discrete_outcomes()]
            self._utility_function = MappingUtilityFunction(
                dict(zip(outcomes, np.random.rand(len(outcomes)))),
            )
        return True


class LimitedOutcomesNegotiator(LimitedOutcomesMixin, SAONegotiator):
    """
    A negotiation agent that uses a fixed set of outcomes in a single
    negotiation.

    Args:
        acceptable_outcomes: the set of acceptable outcomes. If None then it is assumed to be all the outcomes of
                             the negotiation.
        acceptance_probabilities: probability of accepting each acceptable outcome. If None then it is assumed to
                                  be unity.
        proposable_outcomes: the set of outcomes from which the agent is allowed to propose. If None, then it is
                             the same as acceptable outcomes with nonzero probability
        p_no_response: probability of refusing to respond to offers
        p_ending: probability of ending negotiation

    Remarks:
        - The ufun inputs to the constructor and join are ignored. A ufun will be generated that gives a utility equal to
          the probability of choosing a given outcome.
        - If `proposable_outcomes` is passed as None, it is considered the same as `acceptable_outcomes`

    """

    def __init__(
        self,
        acceptable_outcomes: Optional[List["Outcome"]] = None,
        acceptance_probabilities: Optional[Union[float, List[float]]] = None,
        proposable_outcomes: Optional[List["Outcome"]] = None,
        p_ending=0.0,
        p_no_response=0.0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if proposable_outcomes is None:
            proposable_outcomes = acceptable_outcomes
        self.init_limited_outcomes(
            p_ending=p_ending,
            p_no_response=p_no_response,
            acceptable_outcomes=acceptable_outcomes,
            acceptance_probabilities=acceptance_probabilities,
            proposable_outcomes=proposable_outcomes,
        )


# noinspection PyCallByClass
class LimitedOutcomesAcceptor(LimitedOutcomesAcceptorMixin, SAONegotiator):
    """A negotiation agent that uses a fixed set of outcomes in a single
    negotiation.

    Remarks:
        - The ufun inputs to the constructor and join are ignored. A ufun will be generated that gives a utility equal to
          the probability of choosing a given outcome.

    """

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        return LimitedOutcomesAcceptorMixin.respond(self, state=state, offer=offer)

    def __init__(
        self,
        acceptable_outcomes: Optional[List["Outcome"]] = None,
        acceptance_probabilities: Optional[List[float]] = None,
        p_ending=0.0,
        p_no_response=0.0,
        ufun=None,
        **kwargs,
    ) -> None:
        super().__init__(self, **kwargs)
        self.init_limited_outcomes_acceptor(
            p_ending=p_ending,
            p_no_response=p_no_response,
            acceptable_outcomes=acceptable_outcomes,
            acceptance_probabilities=acceptance_probabilities,
        )
        self.add_capabilities({"propose": False})

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        """Always refuses to propose"""
        return None


class AspirationNegotiator(SAONegotiator, AspirationMixin):
    """
    Represents a time-based negotiation strategy that is independent of the offers received during the negotiation.

    Args:
        name: The agent name
        ufun:  The utility function to attache with the agent
        max_aspiration: The aspiration level to use for the first offer (or first acceptance decision).
        aspiration_type: The polynomial aspiration curve type. Here you can pass the exponent as a real value or
                         pass a string giving one of the predefined types: linear, conceder, boulware.
        dynamic_ufun: If True, the utility function will be assumed to be changing over time. This is depricated.
        randomize_offer: If True, the agent will propose outcomes with utility >= the current aspiration level not
                         outcomes just above it.
        can_propose: If True, the agent is allowed to propose
        assume_normalized: If True, the ufun will just be assumed to have the range [0, 1] inclusive
        ranking: If True, the aspiration level will not be based on the utility value but the ranking of the outcome
                 within the presorted list. It is only effective when presort is set to True
        ufun_max: The maximum utility value (used only when `presort` is True)
        ufun_min: The minimum utility value (used only when `presort` is True)
        presort: If True, the negotiator will catch a list of outcomes, presort them and only use them for offers
                 and responses. This is much faster then other option for general continuous utility functions
                 but with the obvious problem of only exploring a discrete subset of the issue space (Decided by
                 the `discrete_outcomes` property of the `AgentMechanismInterface` . If the number of outcomes is
                 very large (i.e. > 10000) and discrete, presort will be forced to be True. You can check if
                 presorting is active in realtime by checking the "presorted" attribute.
        tolerance: A tolerance used for sampling of outcomes when `presort` is set to False
        assume_normalized: If true, the negotiator can assume that the ufun is normalized.
        rational_proposal: If `True`, the negotiator will never propose something with a utility value less than its
                        reserved value. If `propose` returned such an outcome, a NO_OFFER will be returned instead.
        owner: The `Agent` that owns the negotiator.
        parent: The parent which should be an `SAOController`

    """

    def __init__(
        self,
        max_aspiration=1.0,
        aspiration_type="boulware",
        dynamic_ufun=True,
        randomize_offer=False,
        can_propose=True,
        assume_normalized=False,
        ranking=False,
        ufun_max=None,
        ufun_min=None,
        presort: bool = True,
        tolerance: float = 0.01,
        **kwargs,
    ):
        self.ordered_outcomes = []
        self.ufun_max = ufun_max
        self.ufun_min = ufun_min
        self.ranking = ranking
        self.tolerance = tolerance
        if assume_normalized:
            self.ufun_max, self.ufun_min = 1.0, 0.0
        super().__init__(
            assume_normalized=assume_normalized, **kwargs,
        )
        self.aspiration_init(
            max_aspiration=max_aspiration, aspiration_type=aspiration_type
        )
        if not dynamic_ufun:
            warnings.warn(
                "dynamic_ufun is deprecated. All Aspiration negotiators assume a dynamic ufun"
            )
        self.randomize_offer = randomize_offer
        self._max_aspiration = self.max_aspiration
        self.best_outcome, self.worst_outcome = None, None
        self.presort = presort
        self.presorted = False
        self.add_capabilities(
            {
                "respond": True,
                "propose": can_propose,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )
        self.__last_offer_util, self.__last_offer = float("inf"), None
        self.n_outcomes_to_force_presort = 10000
        self.n_trials = 1

    def on_ufun_changed(self):
        super().on_ufun_changed()
        presort = self.presort
        if (
            not presort
            and all(i.is_countable() for i in self._ami.issues)
            and Issue.num_outcomes(self._ami.issues) >= self.n_outcomes_to_force_presort
        ):
            presort = True
        if presort:
            outcomes = self._ami.discrete_outcomes()
            uvals = self.utility_function.eval_all(outcomes)
            uvals_outcomes = [
                (u, o)
                for u, o in zip(uvals, outcomes)
                if u >= self.utility_function.reserved_value
            ]
            self.ordered_outcomes = sorted(
                uvals_outcomes,
                key=lambda x: float(x[0]) if x[0] is not None else float("-inf"),
                reverse=True,
            )
            if self.assume_normalized:
                self.ufun_min, self.ufun_max = 0.0, 1.0
            elif len(self.ordered_outcomes) < 1:
                self.ufun_max = self.ufun_min = self.utility_function.reserved_value
            else:
                if self.ufun_max is None:
                    self.ufun_max = self.ordered_outcomes[0][0]

                if self.ufun_min is None:
                    # we set the minimum utility to the minimum finite value above both reserved_value
                    for j in range(len(self.ordered_outcomes) - 1, -1, -1):
                        self.ufun_min = self.ordered_outcomes[j][0]
                        if self.ufun_min is not None and self.ufun_min > float("-inf"):
                            break
                    if (
                        self.ufun_min is not None
                        and self.ufun_min < self.reserved_value
                    ):
                        self.ufun_min = self.reserved_value
        else:
            if (
                self.ufun_min is None
                or self.ufun_max is None
                or self.best_outcome is None
                or self.worst_outcome is None
            ):
                mn, mx, self.worst_outcome, self.best_outcome = utility_range(
                    self.ufun, return_outcomes=True, issues=self._ami.issues
                )
                if self.ufun_min is None:
                    self.ufun_min = mn
                if self.ufun_max is None:
                    self.ufun_max = mx

        if self.ufun_min < self.reserved_value:
            self.ufun_min = self.reserved_value
        if self.ufun_max < self.ufun_min:
            self.ufun_max = self.ufun_min

        self.presorted = presort
        self.n_trials = 10

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        if self.ufun_max is None or self.ufun_min is None:
            self.on_ufun_changed()
        if self._utility_function is None:
            return ResponseType.REJECT_OFFER
        u = self._utility_function(offer)
        if u is None or u < self.reserved_value:
            return ResponseType.REJECT_OFFER
        asp = (
            self.aspiration(state.relative_time) * (self.ufun_max - self.ufun_min)
            + self.ufun_min
        )
        if u >= asp and u > self.reserved_value:
            return ResponseType.ACCEPT_OFFER
        if asp < self.reserved_value:
            return ResponseType.END_NEGOTIATION
        return ResponseType.REJECT_OFFER

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        if self.ufun_max is None or self.ufun_min is None:
            self.on_ufun_changed()
        if self.ufun_max < self.reserved_value:
            return None
        asp = (
            self.aspiration(state.relative_time) * (self.ufun_max - self.ufun_min)
            + self.ufun_min
        )
        if asp < self.reserved_value:
            return None
        if self.presorted:
            if len(self.ordered_outcomes) < 1:
                return None
            for i, (u, o) in enumerate(self.ordered_outcomes):
                if u is None:
                    continue
                if u < asp:
                    if u < self.reserved_value:
                        return None
                    if i == 0:
                        return self.ordered_outcomes[i][1]
                    if self.randomize_offer:
                        return random.sample(self.ordered_outcomes[:i], 1)[0][1]
                    return self.ordered_outcomes[i - 1][1]
            if self.randomize_offer:
                return random.sample(self.ordered_outcomes, 1)[0][1]
            return self.ordered_outcomes[-1][1]
        else:
            if asp >= 0.99999999999 and self.best_outcome is not None:
                return self.best_outcome
            if self.randomize_offer:
                return outcome_with_utility(
                    ufun=self._utility_function,
                    rng=(asp, float("inf")),
                    issues=self._ami.issues,
                )
            tol = self.tolerance
            for _ in range(self.n_trials):
                rng = self.ufun_max - self.ufun_min
                mx = min(asp + tol * rng, self.__last_offer_util)
                outcome = outcome_with_utility(
                    ufun=self._utility_function, rng=(asp, mx), issues=self._ami.issues,
                )
                if outcome is not None:
                    break
                tol = math.sqrt(tol)
            else:
                outcome = (
                    self.best_outcome
                    if self.__last_offer is None
                    else self.__last_offer
                )
            self.__last_offer_util = self.utility_function(outcome)
            self.__last_offer = outcome
            return outcome


class NiceNegotiator(SAONegotiator, RandomProposalMixin):
    """
    Offers and accepts anything.

    Args:
         name: Negotiator name
         parent: Parent controller if any
         ufun: The ufun of the negotiator
         assume_normalized: If true, the negotiator can assume that the ufun is normalized.
         rational_proposal: If `True`, the negotiator will never propose something with a utility value less than its
                            reserved value. If `propose` returned such an outcome, a NO_OFFER will be returned instead.
         owner: The `Agent` that owns the negotiator.

    """

    def __init__(self, *args, **kwargs):
        SAONegotiator.__init__(self, *args, **kwargs)
        self.init_random_proposal()

    def respond_(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        return ResponseType.ACCEPT_OFFER

    def propose_(self, state: MechanismState) -> Optional["Outcome"]:
        return RandomProposalMixin.propose(self, state)

    propose = propose_


class ToughNegotiator(SAONegotiator):
    """
    Accepts and proposes only the top offer (i.e. the one with highest utility).

    Args:
         name: Negotiator name
         parent: Parent controller if any
         dynamic_ufun: If `True`, assumes a dynamic ufun that can change during the negotiation
         can_propose: If `False` the negotiator will never propose but can only accept
         ufun: The ufun of the negotiator
         rational_proposal: If `True`, the negotiator will never propose something with a utility value less than its
                            reserved value. If `propose` returned such an outcome, a NO_OFFER will be returned instead.
         owner: The `Agent` that owns the negotiator.

    Remarks:
        - If there are multiple outcome with the same maximum utility, only one of them will be used.

    """

    def __init__(
        self,
        name=None,
        parent: Controller = None,
        dynamic_ufun=True,
        can_propose=True,
        ufun=None,
        **kwargs,
    ):
        super().__init__(name=name, parent=parent, ufun=ufun, **kwargs)
        self.best_outcome = None
        self._offerable_outcomes = None
        if not dynamic_ufun:
            warnings.warn(
                "dynamic_ufun is deprecated. All Aspiration negotiators assume a dynamic ufun"
            )
        self.add_capabilities(
            {
                "respond": True,
                "propose": can_propose,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )

    def on_ufun_changed(self):
        super().on_ufun_changed()
        if self._utility_function is None:
            return
        _, _, _, self.best_outcome = utility_range(
            self.utility_function, issues=self.ami.issues, return_outcomes=True,
        )

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        if offer == self.best_outcome:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        if not self._capabilities["propose"]:
            return None
        return self.best_outcome


class OnlyBestNegotiator(SAONegotiator):
    """
    Offers and accepts only one of the top outcomes for the negotiator.

    Args:
         name: Negotiator name
         parent: Parent controller if any
         dynamic_ufun: If `True`, assumes a dynamic ufun that can change during the negotiation
         can_propose: If `False` the negotiator will never propose but can only accept
         ufun: The ufun of the negotiator
         min_utility: The minimum utility to offer or accept
         top_fraction: The fraction of the outcomes (ordered decreasingly by utility) to offer or accept
         best_first: Guarantee offering will non-increasing in terms of utility value
         probabilistic_offering: Offer randomly from the outcomes selected based on `top_fraction` and `min_utility`
         rational_proposal: If `True`, the negotiator will never propose something with a utility value less than its
                            reserved value. If `propose` returned such an outcome, a NO_OFFER will be returned instead.
         owner: The `Agent` that owns the negotiator.
    """

    def __init__(
        self,
        name=None,
        parent: Controller = None,
        dynamic_ufun=True,
        min_utility=0.95,
        top_fraction=0.05,
        best_first=False,
        probabilistic_offering=True,
        can_propose=True,
        ufun=None,
        **kwargs,
    ):
        self._offerable_outcomes = None
        self.best_outcome = []
        self.ordered_outcomes = []
        self.acceptable_outcomes = []
        self.wheel = np.array([])
        self.offered = set([])
        super().__init__(name=name, parent=parent, ufun=ufun, **kwargs)
        if not dynamic_ufun:
            warnings.warn(
                "dynamic_ufun is deprecated. All Aspiration negotiators assume a dynamic ufun"
            )
        self.top_fraction = top_fraction
        self.min_utility = min_utility
        self.best_first = best_first
        self.probabilistic_offering = probabilistic_offering
        self.add_capabilities(
            {
                "respond": True,
                "propose": can_propose,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )

    def on_ufun_changed(self):
        super().on_ufun_changed()
        outcomes = (
            self._ami.discrete_outcomes()
            if self._offerable_outcomes is None
            else self._offerable_outcomes
        )
        eu_outcome = list(zip(self.utility_function.eval_all(outcomes), outcomes))
        self.ordered_outcomes = sorted(eu_outcome, key=lambda x: x[0], reverse=True)
        if self.min_utility is None:
            selected, selected_utils = [], []
        else:
            util_limit = self.min_utility * self.ordered_outcomes[0][0]
            selected, selected_utils = [], []
            for u, o in self.ordered_outcomes:
                if u >= util_limit:
                    selected.append(o)
                    selected_utils.append(u)
                else:
                    break
        if self.top_fraction is not None:
            frac_limit = max(1, round(self.top_fraction * len(self.ordered_outcomes)))
        else:
            frac_limit = len(outcomes)

        if frac_limit >= len(selected) > 0:
            sum = np.asarray(selected_utils).sum()
            if sum > 0.0:
                selected_utils /= sum
                selected_utils = np.cumsum(selected_utils)
            else:
                selected_utils = np.linspace(0.0, 1.0, len(selected_utils))
            self.acceptable_outcomes, self.wheel = selected, selected_utils
            return
        if frac_limit > 0:
            n_sel = len(selected)
            fsel = [_[1] for _ in self.ordered_outcomes[n_sel:frac_limit]]
            futil = [_[0] for _ in self.ordered_outcomes[n_sel:frac_limit]]
            selected_utils = selected_utils + futil
            sum = np.asarray(selected_utils).sum()
            if sum > 0.0:
                selected_utils /= sum
                selected_utils = np.cumsum(selected_utils)
            else:
                selected_utils = np.linspace(0.0, 1.0, len(selected_utils))
            self.acceptable_outcomes, self.wheel = selected + fsel, selected_utils
            return
        self.acceptable_outcomes, self.wheel = [], []
        return

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        if offer in self.acceptable_outcomes:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        if not self._capabilities["propose"]:
            return None
        if self.best_first:
            for o in self.acceptable_outcomes:
                if o not in self.offered:
                    self.offered.add(o)
                    return o
        if len(self.acceptable_outcomes) > 0:
            if self.probabilistic_offering:
                r = random.random()
                for o, w in zip(self.acceptable_outcomes, self.wheel):
                    if w > r:
                        return o
                return random.sample(self.acceptable_outcomes, 1)[0]
            return random.sample(self.acceptable_outcomes, 1)[0]
        return None


class NaiveTitForTatNegotiator(SAONegotiator):
    """
    Implements a naive tit-for-tat strategy that does not depend on the availability of an opponent model.

    Args:
        name: Negotiator name
        ufun: negotiator ufun
        parent: A controller
        kindness: How 'kind' is the agent. A value of zero is standard tit-for-tat. Positive values makes the negotiator
                  concede faster and negative values slower.
        randomize_offer: If `True`, the offers will be randomized above the level determined by the current concession
                        which in turn reflects the opponent's concession.
        always_concede: If `True` the agent will never use a negative concession rate
        initial_concession: How much should the agent concede in the beginning in terms of utility. Should be a number
                            or the special string value 'min' for minimum concession

    Remarks:
        - This negotiator does not keep an opponent model. It thinks only in terms of changes in its own utility.
          If the opponent's last offer was better for the negotiator compared with the one before it, it considers
          that the opponent has conceded by the difference. This means that it implicitly assumes a zero-sum
          situation.
    """

    def __init__(
        self,
        name: str = None,
        parent: Controller = None,
        ufun: Optional["UtilityFunction"] = None,
        kindness=0.0,
        randomize_offer=False,
        always_concede=True,
        initial_concession: Union[float, str] = "min",
        **kwargs,
    ):
        self.received_utilities = []
        self.proposed_utility = None
        self.ordered_outcomes = None
        self.sent_offer_index = None
        self.n_sent = 0
        super().__init__(name=name, ufun=ufun, parent=parent, **kwargs)
        self.kindness = kindness
        self.initial_concession = initial_concession
        self.randomize_offer = randomize_offer
        self.always_concede = always_concede

    def on_ufun_changed(self):
        super().on_ufun_changed()
        outcomes = self._ami.discrete_outcomes()
        self.ordered_outcomes = sorted(
            [(self._utility_function(outcome), outcome) for outcome in outcomes],
            key=lambda x: x[0],
            reverse=True,
        )

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        if self._utility_function is None:
            return ResponseType.REJECT_OFFER
        offered_utility = self._utility_function(offer)
        if len(self.received_utilities) < 2:
            self.received_utilities.append(offered_utility)
        else:
            self.received_utilities[0] = self.received_utilities[1]
            self.received_utilities[-1] = offered_utility
        indx = self._propose(state=state)
        my_utility, my_offer = self.ordered_outcomes[indx]
        if offered_utility >= my_utility:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def _outcome_just_below(self, ulevel: float) -> int:
        for i, (u, o) in enumerate(self.ordered_outcomes):
            if u is None:
                continue
            if u < ulevel:
                if self.randomize_offer:
                    return random.randint(0, i)
                return i
        if self.randomize_offer:
            return random.randint(0, len(self.ordered_outcomes) - 1)
        return -1

    def _propose(self, state: MechanismState) -> int:
        if self.proposed_utility is None:
            return 0
        if len(self.received_utilities) < 2:
            if (
                isinstance(self.initial_concession, str)
                and self.initial_concession == "min"
            ):
                return self._outcome_just_below(ulevel=self.ordered_outcomes[0][0])
            else:
                asp = self.ordered_outcomes[0][0] * (1.0 - self.initial_concession)
            return self._outcome_just_below(ulevel=asp)

        if self.always_concede:
            opponent_concession = max(
                0.0, self.received_utilities[1] - self.received_utilities[0]
            )
        else:
            opponent_concession = (
                self.received_utilities[1] - self.received_utilities[0]
            )
        indx = self._outcome_just_below(
            ulevel=self.proposed_utility
            - opponent_concession
            - self.kindness * max(0.0, opponent_concession)
        )
        return indx

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        indx = self._propose(state)
        self.proposed_utility = self.ordered_outcomes[indx][0]
        return self.ordered_outcomes[indx][1]


class PassThroughSAONegotiator(SAONegotiator):
    """
    A negotiator that acts as an end point to a parent Controller.

    This negotiator simply calls its controler for everything.
    """

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        """Calls parent controller"""
        if self._Negotiator__parent:
            return self._Negotiator__parent.propose(self.id, state)

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        """Calls parent controller"""
        if self._Negotiator__parent:
            return self._Negotiator__parent.respond(self.id, state, offer)
        return ResponseType.REJECT_OFFER

    def on_negotiation_start(self, state: MechanismState) -> None:
        """Calls parent controller"""
        if self._Negotiator__parent:
            return self._Negotiator__parent.on_negotiation_start(self.id, state)

    def on_negotiation_end(self, state: MechanismState) -> None:
        """Calls parent controller"""
        if self._Negotiator__parent:
            return self._Negotiator__parent.on_negotiation_end(self.id, state)

    def join(self, ami, state, *, ufun=None, role="agent",) -> bool:
        """
        Joins a negotiation.

        Remarks:
            This method first gets permission from the parent controller by calling `before_join` on it and confirming
            the result is `True`, it then joins the negotiation and calls `after_join` of the controller to inform it
            that joining is completed if joining was successful.
        """
        permission = (
            self._Negotiator__parent is None
            or self._Negotiator__parent.before_join(
                self.id, ami, state, ufun=ufun, role=role
            )
        )
        if not permission:
            return False
        if super().join(ami, state, ufun=ufun, role=role):
            if self._Negotiator__parent:
                self._Negotiator__parent.after_join(
                    self.id, ami, state, ufun=ufun, role=role
                )
            return True
        return False


def _to_java_response(response: ResponseType) -> int:
    if response == ResponseType.ACCEPT_OFFER:
        return 0
    if response == ResponseType.REJECT_OFFER:
        return 1
    if response == ResponseType.END_NEGOTIATION:
        return 2
    if response == ResponseType.NO_RESPONSE:
        return 3
    raise ValueError(f"Unknown response{response}")


def _from_java_response(response: int) -> ResponseType:
    if response == 0:
        return ResponseType.ACCEPT_OFFER
    if response == 1:
        return ResponseType.REJECT_OFFER
    if response == 2:
        return ResponseType.END_NEGOTIATION
    if response == 3:
        return ResponseType.NO_RESPONSE
    raise ValueError(
        f"Unknown response type {response} returned from the Java underlying negotiator"
    )


class JavaSAONegotiator(SAONegotiator, JavaCallerMixin):
    """
    Represents a negotiator running on JNegMAS (depricated)


    Args:
        java_object: The Java object if known
        java_class_name: Name of the java negotiator class
        auto_load_java: If true, jnemgas will be loaded if not running.
        outcome_type: Type of outcome to use

    Remarks:
        - You cannot pass a ufun here.
    """

    def __init__(
        self,
        java_object,
        java_class_name: Optional[str],
        auto_load_java: bool = False,
        outcome_type: Type = dict,
    ):
        if java_class_name is not None or java_object is not None:
            self.init_java_bridge(
                java_object=java_object,
                java_class_name=java_class_name,
                auto_load_java=auto_load_java,
                python_shadow_object=None,
            )
            if java_object is None:
                self._java_object.fromMap(to_java(self))
        d = {
            python_identifier(k): v
            for k, v in JNegmasGateway.gateway.entry_point.toMap(
                self._java_object
            ).items()
        }
        ufun = d.get("utility_function", None)
        ufun = JavaUtilityFunction(ufun, None) if ufun is not None else None
        super().__init__(
            name=d.get("name", None),
            assume_normalized=d.get("assume_normalized", False),
            ufun=ufun,
            rational_proposal=d.get("rational_proposal", True),
            parent=d.get("parent", None),
        )
        self._outcome_type = outcome_type
        self.add_capabilities(
            {
                "respond": True,
                "propose": True,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )

    def on_partner_proposal(
        self, state: MechanismState, partner_id: str, offer: "Outcome"
    ) -> None:
        self._java_object.onPartnerProposal(to_java(state), partner_id, to_java(offer))

    def on_partner_refused_to_propose(
        self, state: MechanismState, agent_id: str
    ) -> None:
        self._java_object.onPartnerRefusedToPropose(to_java(state), agent_id)

    def on_partner_response(
        self,
        state: MechanismState,
        partner_id: str,
        outcome: "Outcome",
        response: "ResponseType",
    ) -> None:
        self._java_object.onPartnerResponse(
            to_java(state),
            partner_id,
            to_java(outcome),
            self._to_java_response(response),
        )

    def isin(self, negotiation_id: Optional[str]) -> bool:
        return self._java_object.isIn(negotiation_id)

    def join(
        self,
        ami: AgentMechanismInterface,
        state: MechanismState,
        *,
        ufun: Optional["UtilityFunction"] = None,
        role: str = "agent",
    ) -> bool:
        return self._java_object.join(
            java_link(_ShadowAgentMechanismInterface(ami)), to_java(state), ufun, role
        )

    def on_negotiation_start(self, state: MechanismState) -> None:
        self._java_object.onNegotiationStart(to_java(state))

    def on_round_start(self, state: MechanismState) -> None:
        self._java_object.onRoundStart(to_java(state))

    def on_mechanism_error(self, state: MechanismState) -> None:
        self._java_object.onMechanismError(to_java(state))

    def on_round_end(self, state: MechanismState) -> None:
        self._java_object.onRoundEnd(to_java(state))

    def on_leave(self, state: MechanismState) -> None:
        self._java_object.onLeave(to_java(state))

    def on_negotiation_end(self, state: MechanismState) -> None:
        self._java_object.onNegotiationEnd(to_java(state))

    def on_ufun_changed(self):
        self._java_object.onUfunChanged()

    @classmethod
    def from_dict(
        cls, java_object, *args, parent: Controller = None
    ) -> "JavaSAONegotiator":
        """Creates a Java negotiator from an object returned from the JVM implementing PySAONegotiator"""
        ufun = java_object.getUtilityFunction()
        if ufun is not None:
            ufun = JavaUtilityFunction.from_dict(java_object=ufun)
        return JavaCallerMixin.from_dict(
            java_object,
            name=java_object.getName(),
            assume_normalized=java_object.getAssumeNormalized(),
            rational_proposal=java_object.getRationalProposal(),
            parent=parent,
            ufun=ufun,
        )

    def on_notification(self, notification: Notification, notifier: str):
        super().on_notification(notification=notification, notifier=notifier)
        jnotification = {"type": notification.type, "data": to_java(notification.data)}
        self._java_object.on_notification(jnotification, notifier)

    def respond(self, state: MechanismState, offer: "Outcome"):
        return _from_java_response(
            self._java_object.respond(to_java(state), outcome_as_dict(offer))
        )

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        outcome = from_java(self._java_object.propose(to_java(state)))
        if outcome is None:
            return None
        if self._outcome_type == dict:
            return outcome
        if self._outcome_type == tuple:
            return tuple(outcome.values())
        return self._outcome_type(outcome)

    # class Java:
    #    implements = ['jnegmas.sao.SAONegotiator']


class _ShadowSAONegotiator:
    """A python shadow to a java negotiator"""

    class Java:
        implements = ["jnegmas.sao.SAONegotiator"]

    def to_java(self):
        return to_dict(self.shadow)

    def __init__(self, negotiator: SAONegotiator):
        self.shadow = negotiator

    def respond(self, state, outcome):
        return _to_java_response(
            self.shadow.respond(from_java(state), from_java(outcome))
        )

    def propose(self, state):
        return to_java(self.shadow.propose(from_java(state)))

    def isIn(self, negotiation_id):
        return to_java(self.shadow.isin(negotiation_id=negotiation_id))

    def join(self, ami, state, ufun, role):
        return to_java(
            self.shadow.join(
                ami=from_java(ami),
                state=from_java(state),
                ufun=JavaUtilityFunction(ufun, None) if ufun is not None else None,
                role=role,
            )
        )

    def onNegotiationStart(self, state):
        return to_java(self.shadow.on_negotiation_start(from_java(state)))

    def onNegotiationEnd(self, state):
        return to_java(self.shadow.on_negotiation_end(from_java(state)))

    def onMechanismError(self, state):
        return to_java(self.shadow.on_mechanism_error(from_java(state)))

    def onRoundStart(self, state):
        return to_java(self.shadow.on_round_start(from_java(state)))

    def onRoundEnd(self, state):
        return to_java(self.shadow.on_round_end(from_java(state)))

    def onLeave(self, state):
        return to_java(self.shadow.on_leave(from_java(state)))

    def onPartnerProposal(self, state, agent_id, offer):
        return to_java(
            self.shadow.on_partner_proposal(
                state=from_java(state), partner_id=agent_id, offer=from_java(offer)
            )
        )

    def onUfunChanged(self):
        return to_java(self.shadow.on_ufun_changed())

    def onPartnerRefusedToPropose(self, state, agent_id):
        return to_java(
            self.shadow.on_partner_refused_to_propose(
                state=from_java(state), agent_id=agent_id
            )
        )

    def onPartnerResponse(self, state, agent_id, offer, response: int, counter_offer):
        # TODO check that jnemgas is expecting ResponseType not SAOResponse here.
        return to_java(
            self.shadow.on_partner_response(
                state=from_java(state),
                partner_id=agent_id,
                outcome=from_java(offer),
                response=ResponseType(response),
            )
        )

    def onNotification(self, notification, notifier):
        return to_java(self.shadow.on_notification(from_java(notification), notifier))

    def setUtilityFunction(self, ufun):
        self.shadow.utility_function = (
            ufun if ufun is None else JavaUtilityFunction(ufun, None)
        )

    def getUtilityFunction(self):
        return to_java(self.shadow._utility_function)

    def getID(self):
        return self.shadow.id

    def setID(self):
        return self.shadow.id


SimpleTitForTatNegotiator = NaiveTitForTatNegotiator
"""A simple tit-for-tat negotiator"""



from typing import Type

import dash
import dash_core_components as dcc
import dash_daq as daq
import dash_html_components as html
import dash_table as tbl
from dash.dependencies import Input, Output, State

from negmas.common import NamedObject
from negmas.visualizers import visualizer_type


def layout(object_type: Type[NamedObject]):
    """Returns a layout appropriate for the given named object"""

    layout = [html.Div(id="basic_info", children=None)]
    v = visualizer_type(object_type)
    children = v.children
    if len(children) > 0:
        c = html.Div(id="children")
        for category, v in children:
            c.append(html.H1(category))
            c.append(html.Div(id=category))



"""Defines import/export functionality
"""
import functools
import operator
import os
import shutil
import xml.etree.ElementTree as ET
from os import listdir
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pkg_resources

from .generics import ivalues
from .negotiators import Negotiator
from .outcomes import Issue, enumerate_outcomes
from .sao import AspirationNegotiator, SAOMechanism
from .utilities import UtilityFunction, make_discounted_ufun

__all__ = [
    "load_genius_domain",
    "load_genius_domain_from_folder",
    "convert_genius_domain_from_folder",
    "convert_genius_domain",
    "find_domain_and_utility_files",
    "get_domain_issues",
]


def get_domain_issues(
    domain_file_name: str,
    force_single_issue=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    safe_parsing=False,
) -> Union[Dict[str, Issue], List[Issue]]:
    """
    Returns the issues of a given XML domain (Genius Format)

    Args:
        domain_file_name:
        force_single_issue:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        safe_parsing:

    Returns:
        List or Dict of issues

    """
    issues, issues_details, mechanism = None, None, None
    if domain_file_name is not None:
        domain_file_name = str(domain_file_name)
        issues_details, _ = Issue.from_genius(
            domain_file_name,
            force_single_issue=False,
            keep_issue_names=True,
            keep_value_names=True,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
        if force_single_issue:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                max_n_outcomes=max_n_outcomes,
                n_discretization=n_discretization,
            )
            if issues is None:
                return []
        else:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                safe_parsing=safe_parsing,
                n_discretization=n_discretization,
            )
    return issues if not force_single_issue else [issues]


def load_genius_domain(
    domain_file_name: str,
    utility_file_names: Optional[List[str]] = None,
    agent_factories: Optional[
        Union[Callable[[], Negotiator], List[Callable[[], Negotiator]]]
    ] = None,
    force_single_issue=False,
    force_numeric=False,
    cache_and_discretize_outcomes=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    normalize_utilities=True,
    n_steps=None,
    time_limit=3 * 60,  # GENIUS uses 3min time limit by default
    max_n_agents=None,
    dynamic_entry=True,
    safe_parsing=False,
    ignore_reserved=False,
    ignore_discount=False,
) -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
    """
    Loads a genius domain, creates appropriate negotiators if necessary

    Args:
        domain_file_name:
        utility_file_names:
        agent_factories:
        force_single_issue:
        cache_and_discretize_outcomes:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        normalize_utilities:
        n_steps:
        time_limit:
        max_n_agents:
        dynamic_entry:
        safe_parsing:
        ignore_reserved:
        ignore_discount:

    Returns:
        - mechanism (SAOMechanism): A mechanism for the given issues
        - agent_info (List[Dict]): All Negotiator functions from the given file
        - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues

    """
    issues, issues_details, mechanism = None, None, None
    if domain_file_name is not None:
        domain_file_name = str(domain_file_name)
        issues_details, _ = Issue.from_genius(
            domain_file_name,
            force_single_issue=False,
            keep_issue_names=True,
            keep_value_names=True,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
        if force_single_issue:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                max_n_outcomes=max_n_outcomes,
                n_discretization=n_discretization,
                force_numeric=force_numeric,
            )
            if issues is None:
                return None, [], []
        else:
            issues, _ = Issue.from_genius(
                domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                safe_parsing=safe_parsing,
                n_discretization=n_discretization,
                force_numeric=force_numeric,
            )

    agent_info = []
    if utility_file_names is None:
        utility_file_names = []
    utility_file_names = [str(_) for _ in utility_file_names]
    for ufname in utility_file_names:
        utility, discount_factor = UtilityFunction.from_genius(
            file_name=ufname,
            force_single_issue=force_single_issue,
            keep_issue_names=keep_issue_names,
            keep_value_names=keep_value_names,
            normalize_utility=normalize_utilities,
            domain_issues=issues_details,
            safe_parsing=safe_parsing,
            max_n_outcomes=max_n_outcomes,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            force_numeric=force_numeric,
        )
        agent_info.append(
            {
                "ufun": utility,
                "ufun_name": ufname,
                "reserved_value_func": utility.reserved_value
                if utility is not None
                else 0.0,
                "discount_factor": discount_factor,
            }
        )
    outcomes = None
    try:
        if force_single_issue or cache_and_discretize_outcomes or len(issues) == 1:
            n_outcomes: float = functools.reduce(
                operator.mul,
                (
                    float(_.cardinality)
                    if not _.is_uncountable()
                    else float(n_discretization)
                    if n_discretization is not None
                    else np.inf
                    for _ in ivalues(issues)
                ),
                1.0,
            )
            if n_outcomes < max_n_outcomes:
                outcomes = enumerate_outcomes(
                    issues, astype=dict if keep_issue_names else tuple
                )
    except ValueError:
        pass
    if domain_file_name is not None:
        mechanism_name = (
            domain_file_name.split("/")[-1][:-4]
            .replace("-domain", "")
            .replace("_domain", "")
            .replace("domain", "")
        )
        mechanism = SAOMechanism(
            issues=issues,
            outcomes=outcomes,
            n_steps=n_steps,
            time_limit=time_limit,
            max_n_agents=max_n_agents,
            dynamic_entry=dynamic_entry,
            name=mechanism_name,
            outcome_type=dict if keep_issue_names else tuple,
            enable_callbacks=True,
        )
        if agent_info is not None and len(agent_info) > 0:
            for info in agent_info:
                info["ufun"] = (
                    info["ufun"]
                    if info["discount_factor"] is None or info["discount_factor"] == 1.0
                    else make_discounted_ufun(
                        ufun=info["ufun"],
                        ami=mechanism.ami,
                        discount_per_round=info["discount_factor"],
                        power_per_round=1.0,
                    )
                )

    if agent_factories is not None and agent_info is not None and len(agent_info) > 0:
        if not isinstance(agent_factories, Iterable):
            agent_factories = [agent_factories] * len(agent_info)
        agents = [factory() for factory in agent_factories[0 : len(agent_info)]]
        for a, info in zip(agents, agent_info):
            mechanism.add(a, ufun=info["ufun"])

    return mechanism, agent_info, (issues if not force_single_issue else [issues])


def load_genius_domain_from_folder(
    folder_name: str,
    agent_factories: Optional[
        Union[Callable[[], Negotiator], List[Callable[[], Negotiator]]]
    ] = None,
    force_single_issue=False,
    force_numeric=False,
    cache_and_discretize_outcomes=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    normalize_utilities=True,
    n_steps=None,
    time_limit=60,  # GENIUS uses 3min time limit by default
    max_n_agents=None,
    dynamic_entry=True,
    safe_parsing=False,
    ignore_reserved=False,
    ignore_discount=False,
) -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
    """
    Loads a genius domain from a folder. See ``load_genius_domain`` for more details.

    Args:
        folder_name:
        agent_factories:
        force_single_issue:
        cache_and_discretize_outcomes:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        normalize_utilities:
        n_steps:
        time_limit:
        max_n_agents:
        dynamic_entry:
        safe_parsing:
        ignore_reserved:
        ignore_discount:

    Returns:
        - mechanism (SAOMechanism): A mechanism for the given issues
        - agent_info (List[Dict]): All Negotiator functions from the given file
        - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues

    Examples:

        >>> folder_name = pkg_resources.resource_filename('negmas', resource_name='tests/data/10issues')
        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(folder_name
        ...                             , force_single_issue=False, keep_issue_names=False
        ...                             , keep_value_names=False, normalize_utilities=False)
        >>> print(len(issues))
        10
        >>> print(len(negotiators))
        2
        >>> print([type(a['ufun']) for a in negotiators])
        [<class 'negmas.utilities.HyperRectangleUtilityFunction'>, <class 'negmas.utilities.HyperRectangleUtilityFunction'>]
        >>> print(negotiators[0]['ufun'].outcome_ranges[0])
        {1: (7.0, 9.0), 3: (2.0, 7.0), 5: (0.0, 8.0), 8: (0.0, 7.0)}
        >>> print(negotiators[0]['ufun'].mappings[0])
        97.0
        >>> u = negotiators[0]['ufun']
        >>> print(u(tuple([0.0] * len(issues))))
        0.0
        >>> print(u(tuple([0.5] * len(issues))))
        186.0

        Try loading and running a domain with predetermined agents:
        >>> mechanism, agents, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        ...                             , agent_factories=AspirationNegotiator
        ...                             , force_single_issue=True, keep_issue_names=False
        ...                             , keep_value_names=False)
        >>> state = mechanism.run()
        >>> state.agreement is not None
        True


        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop'))

        >>> len(issues), len(negotiators)
        (3, 2)

        >>> [type(a['ufun']) for a in negotiators]
        [<class 'negmas.utilities.LinearUtilityAggregationFunction'>, <class 'negmas.utilities.LinearUtilityAggregationFunction'>]

        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        ...                             , force_single_issue=True, keep_issue_names=False
        ...                             , keep_value_names=False)
        >>> len(issues), len(negotiators)
        (1, 2)
        >>> [type(a['ufun']) for a in negotiators]
        [<class 'negmas.utilities.MappingUtilityFunction'>, <class 'negmas.utilities.MappingUtilityFunction'>]

    """
    folder_name = str(folder_name)
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    return load_genius_domain(
        domain_file_name=domain_file_name,
        utility_file_names=utility_file_names,
        agent_factories=agent_factories,
        force_numeric=force_numeric,
        force_single_issue=force_single_issue,
        cache_and_discretize_outcomes=cache_and_discretize_outcomes,
        max_n_outcomes=max_n_outcomes,
        n_discretization=n_discretization,
        keep_issue_names=keep_issue_names,
        keep_value_names=keep_value_names,
        normalize_utilities=normalize_utilities,
        n_steps=n_steps,
        time_limit=time_limit,
        max_n_agents=max_n_agents,
        dynamic_entry=dynamic_entry,
        safe_parsing=safe_parsing,
        ignore_reserved=ignore_reserved,
        ignore_discount=ignore_discount,
    )


def find_domain_and_utility_files(folder_name) -> Tuple[str, List[str]]:
    """Finds the domain and utility_function files in a folder
    """
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names


def convert_genius_domain(
    src_domain_file_name: str,
    dst_domain_file_name: str,
    src_utility_file_names: Optional[List[str]] = None,
    dst_utility_file_names: Optional[List[str]] = None,
    force_single_issue=False,
    cache_and_discretize_outcomes=False,
    max_n_outcomes: int = 1e6,
    n_discretization: Optional[int] = None,
    keep_issue_names=True,
    keep_value_names=True,
    normalize_utilities=True,
    safe_parsing=False,
) -> bool:
    if (
        not force_single_issue
        and not cache_and_discretize_outcomes
        and keep_issue_names
        and keep_value_names
        and not normalize_utilities
    ):
        # no need to do anything, just copy
        shutil.copy(src=src_domain_file_name, dst=dst_domain_file_name)
        for src, dst in zip(src_utility_file_names, dst_utility_file_names):
            shutil.copy(src=src, dst=dst)
        return True
    issues, issues_details, mechanism = None, None, None
    if src_domain_file_name is not None:
        issues_details, _ = Issue.from_genius(
            src_domain_file_name,
            force_single_issue=False,
            keep_issue_names=True,
            keep_value_names=True,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
        if force_single_issue:
            issues, _ = Issue.from_genius(
                src_domain_file_name,
                force_single_issue=force_single_issue,
                keep_issue_names=keep_issue_names,
                keep_value_names=keep_value_names,
                max_n_outcomes=max_n_outcomes,
                n_discretization=n_discretization,
            )
        else:
            issues = issues_details
        if issues is None:
            return False
        Issue.to_genius(
            issues=issues, file_name=dst_domain_file_name, enumerate_integer=True
        )

    if src_utility_file_names is None:
        src_utility_file_names = []
    for ufname, dstfname in zip(src_utility_file_names, dst_utility_file_names):
        utility, discount_factor = UtilityFunction.from_genius(
            file_name=ufname,
            force_single_issue=force_single_issue,
            keep_issue_names=keep_issue_names,
            keep_value_names=keep_value_names,
            normalize_utility=normalize_utilities,
            domain_issues=issues_details,
            safe_parsing=safe_parsing,
            max_n_outcomes=max_n_outcomes,
        )
        if utility is None:
            return False
        UtilityFunction.to_genius(
            u=utility,
            discount_factor=discount_factor,
            issues=issues,
            file_name=dstfname,
        )

    return True


def convert_genius_domain_from_folder(
    src_folder_name: str, dst_folder_name: str, **kwargs
) -> bool:
    """Loads a genius domain from a folder. See ``load_genius_domain`` for more details.



    """
    os.makedirs(dst_folder_name, exist_ok=True)
    files = sorted(listdir(src_folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = src_folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    success = convert_genius_domain(
        src_domain_file_name=domain_file_name,
        dst_domain_file_name=os.path.join(
            dst_folder_name, os.path.basename(domain_file_name)
        ),
        src_utility_file_names=utility_file_names,
        dst_utility_file_names=[
            os.path.join(dst_folder_name, os.path.basename(_))
            for _ in utility_file_names
        ],
        **kwargs
    )

    if not success:
        if len(os.listdir(dst_folder_name)) == 0:
            shutil.rmtree(dst_folder_name, ignore_errors=True)

    return success



import time

from typing import Optional, Union

from .base import BaseElicitor
from ..common import MechanismState
from ..outcomes import Outcome
from ..utilities import (
    IPUtilityFunction,
    UtilityDistribution,
    UtilityValue,
)

__all__ = ["DummyElicitor", "FullKnowledgeElicitor"]


class DummyElicitor(BaseElicitor):
    """
    A dummy elicitation algorithm that does not do any elicitation.
    """

    def utility_on_rejection(
        self, outcome: "Outcome", state: MechanismState
    ) -> UtilityValue:
        return self.reserved_value

    def can_elicit(self) -> bool:
        return True

    def elicit_single(self, state: MechanismState):
        return False

    def init_elicitation(
        self,
        ufun: Optional[Union["IPUtilityFunction", "UtilityDistribution"]],
        **kwargs,
    ):
        super().init_elicitation(ufun=ufun, **kwargs)
        strt_time = time.perf_counter()
        self.offerable_outcomes = self._ami.outcomes
        self._elicitation_time += time.perf_counter() - strt_time


class FullKnowledgeElicitor(BaseElicitor):
    """
    An elicitor that does not *need* to do any elicitation because it has full access
    to the user ufun.
    """

    def utility_on_rejection(
        self, outcome: "Outcome", state: MechanismState
    ) -> UtilityValue:
        return self.reserved_value

    def can_elicit(self) -> bool:
        return True

    def elicit_single(self, state: MechanismState):
        return False

    def init_elicitation(
        self,
        ufun: Optional[Union["IPUtilityFunction", "UtilityDistribution"]],
        **kwargs,
    ):
        super().init_elicitation(ufun=self.user.ufun)
        strt_time = time.perf_counter()
        self.offerable_outcomes = self._ami.outcomes
        self._elicitation_time += time.perf_counter() - strt_time



"""
Implements basic components that can be used by `SAONegotiator` s.
"""
import random
from collections import defaultdict
from typing import Iterable, List, Optional, Tuple, Union

from ..common import MechanismState
from ..negotiators import Negotiator
from ..outcomes import Outcome, ResponseType, outcome_as_tuple
from ..utilities import MappingUtilityFunction

__all__ = [
    "LimitedOutcomesAcceptorMixin",
    "LimitedOutcomesProposerMixin",
    "LimitedOutcomesMixin",
    "RandomResponseMixin",
    "RandomProposalMixin",
]


class RandomResponseMixin:
    """
    A mixin that adds the ability to respond randomly to offers

    Remarks:
        - Call `init_random_response` to initialize the mixin.
    """

    def init_random_response(
        self,
        p_acceptance: float = 0.15,
        p_rejection: float = 0.25,
        p_ending: float = 0.1,
    ) -> None:
        """
        Initializes the mixin.

        Args:
            p_acceptance: probability of accepting offers
            p_rejection: probability of rejecting offers
            p_ending: probability of ending negotiation

        Remarks:
            - If the summation of acceptance, rejection and ending probabilities
                is less than 1.0 then with the remaining probability a
                NO_RESPONSE is returned from respond()
        """
        self.add_capabilities({"respond": True})
        self.p_acceptance = p_acceptance
        self.p_rejection = p_rejection
        self.p_ending = p_ending
        self.wheel: List[Tuple[float, ResponseType]] = [(0.0, ResponseType.NO_RESPONSE)]
        if self.p_acceptance > 0.0:
            self.wheel.append(
                (self.wheel[-1][0] + self.p_acceptance, ResponseType.ACCEPT_OFFER)
            )
        if self.p_rejection > 0.0:
            self.wheel.append(
                (self.wheel[-1][0] + self.p_rejection, ResponseType.REJECT_OFFER)
            )
        if self.p_ending > 0.0:
            self.wheel.append(
                (self.wheel[-1][0] + self.p_ending, ResponseType.REJECT_OFFER)
            )
        if self.wheel[-1][0] > 1.0:
            raise ValueError("Probabilities of acceptance+rejection+ending>1")

        self.wheel = self.wheel[1:]

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        r = random.random()
        for w in self.wheel:
            if w[0] >= r:
                return w[1]
        return ResponseType.NO_RESPONSE


class RandomProposalMixin:
    """
    A mixin that adds the ability to propose random offers

    Remarks:
        - call `init_random_proposal` to initialize the mixin
    """

    def init_random_proposal(self: Negotiator):
        """Initializes the mixin"""
        self.add_capabilities(
            {
                "propose": True,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        """
        Proposes a random offer

        Args:
            state: The mechanism state

        Returns:
            A proposed outcome or None to refuse to offer.

        Remarks:
            - Does not use the state.
        """
        if (
            hasattr(self, "_offerable_outcomes")
            and self._offerable_outcomes is not None
        ):
            return random.sample(self._offerable_outcomes, 1)[0]
        return self._ami.random_outcomes(1)[0]


class LimitedOutcomesAcceptorMixin:
    """
    A mixin that adds the ability to accept random offers
    """

    def init_limited_outcomes_acceptor(
        self,
        acceptable_outcomes: Optional[Iterable["Outcome"]] = None,
        acceptance_probabilities: Optional[List[float]] = None,
        time_factor: Union[float, List[float]] = None,
        p_ending: float = 0.05,
        p_no_response: float = 0.0,
    ) -> None:
        """Initializes the mixin

        Args:
            acceptable_outcomes: the set of acceptable outcomes. If None then it is assumed to be all the outcomes of the negotiation.
            acceptance_probabilities: probability of accepting each acceptable outcome. If None then it is assumed to be unity.
            time_factor: If given, the acceptance probability will go up with time by this factor
            p_no_response: probability of refusing to respond to offers
            p_ending: probability of ending negotiation

        Returns:
            None
        """
        self.add_capabilities({"respond": True})
        self.acceptable_outcomes, self.acceptance_probabilities = (
            acceptable_outcomes,
            acceptance_probabilities,
        )
        self.p_no_response = p_no_response
        self.p_ending = p_ending + p_no_response
        self.time_factor = time_factor

    def join(self, *args, **kwargs):
        if not super().join(*args, **kwargs):
            return False

        self._make_ufun()
        return True

    def _make_ufun(self):
        """Generates a ufun that maps acceptance probability to the utility"""
        if self.acceptable_outcomes is None:
            self.acceptable_outcomes = self.ami.discrete_outcomes()

        self.acceptable_outcomes = [
            outcome_as_tuple(_) for _ in self.acceptable_outcomes
        ]

        if self.acceptance_probabilities is None:
            self.acceptance_probabilities = [1.0] * len(self.acceptable_outcomes)
        if not isinstance(self.acceptance_probabilities, Iterable):
            self.acceptance_probabilities = [self.acceptance_probabilities] * len(
                self.acceptable_outcomes
            )
        self.mapping = defaultdict(float)
        for p, o in zip(self.acceptance_probabilities, self.acceptable_outcomes):
            self.mapping[o] = p
        self.utility_function = MappingUtilityFunction(self.mapping)
        self.on_ufun_changed()

    def respond(self, state: MechanismState, offer: "Outcome") -> "ResponseType":
        """Respond to an offer.

        Args:
            offer (Outcome): offer being tested

        Returns:
            ResponseType: The response to the offer

        """
        # offer = outcome_as_tuple(offer)
        if not hasattr(self, "mapping"):
            self._make_ufun()
        r = random.random()
        if r < self.p_no_response:
            return ResponseType.NO_RESPONSE

        if r < self.p_ending:
            return ResponseType.END_NEGOTIATION

        if random.random() < self.mapping[outcome_as_tuple(offer)]:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


class LimitedOutcomesProposerMixin:
    """
    A mixin that adds the ability to propose an outcome from a set of outcomes randomly.
    """

    def init_limited_outcomes_proposer(
        self: Negotiator, proposable_outcomes: Optional[List["Outcome"]] = None
    ) -> None:
        """
        Initializes the mixin

        Args:
            proposable_outcomes: the set of prooposable outcomes. If None then it is assumed to be all the outcomes of the negotiation

        """
        self.add_capabilities(
            {
                "propose": True,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )
        self._offerable_outcomes = proposable_outcomes
        if proposable_outcomes is not None:
            self._offerable_outcomes = list(proposable_outcomes)

    def propose(self, state: MechanismState) -> Optional["Outcome"]:
        """Proposes one of the proposable_outcomes"""
        if self._offerable_outcomes is None:
            return self._ami.random_outcomes(1)[0]
        else:
            return random.sample(self._offerable_outcomes, 1)[0]


class LimitedOutcomesMixin(LimitedOutcomesAcceptorMixin, LimitedOutcomesProposerMixin):
    """
    A mixin that adds the ability to propose and respond a limited set of outcomes.
    """

    def init_limited_outcomes(
        self,
        acceptable_outcomes: Optional[Iterable["Outcome"]] = None,
        acceptance_probabilities: Optional[Union[float, List[float]]] = None,
        proposable_outcomes: Optional[Iterable["Outcome"]] = None,
        p_ending=0.0,
        p_no_response=0.0,
    ) -> None:
        """Initializes the mixin.

        Args:
            acceptable_outcomes: the set of acceptable outcomes. If None then it is assumed to be all the outcomes of
                                 the negotiation.
            acceptance_probabilities: probability of accepting each acceptable outcome. If None then it is assumed to
                                      be unity.
            proposable_outcomes: the set of outcomes from which the agent is allowed to propose. If None, then it is
                                 the same as acceptable outcomes with nonzero probability
            p_no_response: probability of refusing to respond to offers
            p_ending: probability of ending negotiation
        """
        self.init_limited_outcomes_acceptor(
            acceptable_outcomes=acceptable_outcomes,
            acceptance_probabilities=acceptance_probabilities,
            p_ending=p_ending,
            p_no_response=p_no_response,
        )

        if proposable_outcomes is None and self.acceptable_outcomes is not None:
            if isinstance(self.acceptance_probabilities, Iterable):
                proposable_outcomes = [
                    _
                    for _, p in zip(
                        self.acceptable_outcomes, self.acceptance_probabilities
                    )
                    if p > 1e-9
                ]
        self.init_limited_outcomes_proposer(proposable_outcomes=proposable_outcomes)
